# ESA (C) 2000-2022
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

# XMMextractor_tools

import os
from astropy.coordinates import SkyCoord
from astropy import units as u
from astropy.coordinates import Angle
from astropy.time import Time
from astropy.io import fits
import subprocess
from pysas.pyutils import pyutils
from logger import get_logger
import glob
import shutil
from pysas.sastask import MyTask as W
import warnings
import tarfile
import urllib.request
from astroquery.esa.xmm_newton import XMMNewton 

tool_log = get_logger('aux.log')

def ODF_times(SAS_file = ''):
    '''
    Searches for the start time and the end time of the observation.
    
    Args:
        SAS_file: (optional) a SAS file. By default it will search for one 
        in the main directory for XMMExtractor.
    
    Output:
        (start time, end time, duration)
    '''

    if SAS_file == '':
        SAS_file = glob.glob(XMMExtractor.main_dir + '/*.SAS')[0]

    t_start = 0
    t_end = 0
    duration = 0

    with open(SAS_file, 'r') as f:
        for line in f:
            if 'Observation Start Time' in line:
                t_start = line.replace(' / Observation Start Time', '')
                t_start = Time(t_start)
            if 'Observation End Time' in line:
                t_end = line.replace(' / Observation End Time', '')
                t_end = Time(t_end)
            if t_start != 0 and t_end != 0:
                break # There is no interested in the rest of the file.
        duration = t_end - t_start

    return (t_start, t_end, duration.value*86400)


def ra(source_name):
    '''
    '''

    cmd = ned_coord_directory + ned_coord_script + source_name
    ps = subprocess.Popen(cmd, shell = True, stdout = subprocess.PIPE, stderr = subprocess.STDOUT)
    out = ps.communicate()[0]

    out = out.split()

    if out[0] == 'Target':
        ra = 0.
    else:
        ra = out[0]

    return ra


def dec(source_name):
    '''
    '''

    cmd = ned_coord_directory + ned_coord_script + source_name
    ps = subprocess.Popen(cmd, shell = True, stdout = subprocess.PIPE, stderr = subprocess.STDOUT)
    out = ps.communicate()[0]

    out = out.split()

    if out[0] == 'Target':
        dec = 0.
    else:
        dec = out[1]

    return dec


def angular_distance(ra1, ra2, dec1, dec2):
    '''
    Returns the angular distance between two objectives in degrees.

    Args:
        ra1: RA of the first item.
        ra2: RA of the second item.
        dec1: DEC of the first item.
        dec2: DEC of the second item.
    
    Output: 
        angular_distance: The angular distance between the objectives, in AstroPy degree units format.
    '''

    # TODO: dheck units (and add them)
    coord1 = SkyCoord(ra1 * u.deg, dec1 * u.deg)
    coord2 = SkyCoord(ra2 * u.deg, dec2 * u.deg)

    angular_distance = coord1.separation(coord2)

    return angular_distance


def add_source_to_region_file(region_file, x, y, r1, r2, color, text, region_type, coord_type):
    '''
    Creates a region file that can be read with ds9.

    Args:
        region_file: the name of the region file.
        x: x coordinate.
        y: y coordinate.
        r1: radius 1 (use r1 = 0 for circle)
        r2: radius 2.
        color: desired colour.
        text: added text.
        region_type: region type.
        coord_type: type of coordinates.

    Output:
        0: if finished; 1 if the region could not be read.
    '''

    with open(region_file, 'w+') as rg_file:
        if region_type == 'circle':
            rg_file.write('{};circle({},{},{}) # color={} text=\"{}\" \n'.format(coord_type, x, y, r2, color, text))
        elif region_type == 'annulus':
            rg_file.write('{};annulus({},{},{},{}) # color={} text=\"{}\" \n'.format(coord_type, x, y, r1, r2, color, text))
        elif region_type == 'box':
            rg_file.write('{};box({},{},{},{}) # color={} text=\"{}\" \n'.format(coord_type, x, y, r1, r2, color, text))
        else:
            warnings.warn('Shape not accepted. Only circle, annulus and box.')
            return 1

    return 0


def produce_ps_file(fits_file, region_file, title, ps_file):
    '''
    Produces a PS file.
    
    Args:
        fits_file: the given FITS file.
        region_file: the name of the region file.
        title: the title of the PS.
        ps_file: the name of the PS file.

    Output:
        0 if run successfully, 1 otherwise.
    '''

    args = 'ds9 -cmap invert yes -cmap b -cmap value 1.15 .13 -fits {} '.format(fits_file) +\
        '-region {} -zoom to fit -title {} -print destination file '.format(region_file, title)+\
        '-print filename {} -print palette rgb -print level 1 -print resolution 300 -print -exit'.format(ps_file)

    if os.system(args):
        return 1

    return 0


def deg_to_hours(deg):
    '''
    Moves from degrees to hours.
    
    Args:
        deg: degrees.
    
    Output: 
        hour: the equivalent in hours.
    '''

    if hasattr(deg, 'unit'):
        if deg.unit == 'hourangle':
            print('Already in hours')
            hour = deg
        else:
            hour = deg.to(u.hourangle)
    else:
        deg = deg * u.deg
        hour = deg.to(u.hourangle)

    return hour


def deg_to_sexagesimal(deg):
    '''
    Moves from degrees to sexagesimal degrees.
    
    Args:
        deg: degrees.
    
    Output:
        dms: the equivalent in dms as a tuple.
    '''

    if hasattr(deg, 'unit'):
        if deg.unit == 'deg':
            dms = Angle(deg.value, deg.unit)
            dms = dms.dms
    else:
        deg = Angle(deg, u.deg)
        dms = deg.dms

    return dms


def event_file_spectral_info(event_file, instr):
    '''
    Writes into the log file some basic information regarding the given event list.
    
    Args:
        event_file: the event list.
        instr: the instrument.

    Output: 
        (exposure live time, backscale, areascale).
    '''

    #instr = PythonUtils.normalize_instrument(instrument)

    #live_time = PythonUtils.get_key_word(event_file, 'EXPOSURE', 1)
    live_time = pyutils.get_key_word(event_file, 'EXPOSURE', 1)
    
    if live_time == 'EXPOSURE' or live_time == '' or live_time == 0:
        live_time = 'unknown'

    backscale = get_back_scale(event_file, 1)
    areascale = get_area_scale(event_file, 1)

    tool_log.info('   #> Information from Spectra file: {}', event_file)
    tool_log.info('   {} live time : {} (Weighted live time of CCDs in extraction region)', instr, live_time)
    tool_log.info('   {} Backscale : {}', instr, backscale)
    tool_log.info('   {} Areascale : {}', instr, areascale)

    return (live_time, backscale, areascale)


def get_back_scale(event_file, extension):
    '''
    Checks for the BACKSCAL in a given event file and extension.
    
    Args:
        event_file: the event file.
        extension: the extension of that event file.
    
    Output: 
        backscale: the backscale.
    '''

    backscale = pyutils.get_key_word(event_file, 'BACKSCAL', extension)
    if backscale == 'BACKSCAL' or backscale == '':
        backscale = 'unknown'
    
    return backscale


def get_area_scale(event_file, extension):
    '''
    Checks for the AREASCAL in a given event file and extension.

    Args:
        event_file: the event file.
        extension: the extension of that event file.
    
    Output: 
        areascale: the areascale.
    '''

    areascale = pyutils.get_key_word(event_file, 'AREASCAL', extension)
    if areascale == 'AREASCAL' or areascale == '':
        areascale = 'unknown'
    
    return areascale


def get_entries(event_file, extension):
    '''
    Checks for the number of entries in a given event file and extension.
    
    Args:
        event_file: the event file.
        extension: the extension of that event file.

    Output:
        n_entries: the number of entries found in the FITS file.
    '''
    
    n_entries = pyutils.get_key_word(event_file, 'NAXIS2', extension)
    return n_entries


def print_SAS_setup():
    '''
    Prints the SAS information regarding env variables into the XMMextractor
    log file.
    '''

    list_var = os.environ
    SAS_info = ''
    for i in list_var:
        if 'SAS' in i:
            SAS_info = SAS_info + i + '=' + os.environ.get(i) + '\n'

    tool_log.debug('   #> This is the current SAS set-up: ')
    tool_log.info(SAS_info)


def set_coordinates(obs, ra = '', dec = ''):
    '''
    For a given observation object, modifies the RA and DEC of it's attributes
    and in the XML file.
    #TODO: check consistency with the GUI.
    
    Args:
        obs: the observation object.
        ra:: the new RA.
        dec: the new DEC.

    Output:
        obs: the observation object with the applied changes.
    '''


    test_odfingest = run_cifbuild.test_odfingest()
    odfingest_file = test_odfingest[1]

    if not test_odfingest[0]:
        tool_log.error('    Run odfingest first')
        RunCifbuild.run_odfingest()

    if ra != '' or dec != '':
        tool_log.info('   #> Using user defined coordinates')
        if ra != '':
            obs.set_coord(ra = str(ra))
        if dec != '':
            obs.set_coord(dec = str(ra))

        # obs.writeXML()
        # filename = glob.glob('*.xml')[0]
        # obs = XMMXMLextractor_classes.Observation(filename)

    else:
        tool_log.info("   #> Not user defined coordinates found. Using the SUM.SAS RA_OBJ and DE_OBJ coordinates.")
        tool_log.info('   #> RA_OBJ = {}', obs.ra)
        tool_log.info('   #> DEC_OBJ = {}', obs.dec)

    return obs


def produce_image(event_file, image_file, instr, mode):
    '''
    Generates an image for an EPN or EMOS1, EMOS2 event list.

    Args:
        event_file: the event file.
        image_file: the name for the image file.
        instr: the instrument (will be normalized)
        mode: the mode

    Output:
        0 if run successfully; 1 otherwise.
    '''

    tool_log.info('   #> Producing image from {} .... ', event_file)
    if instr == 'EPN':
        expr = '#XMMEA_EP'
    if instr == 'EMOS1':
        expr = '#XMMEA_EM'
    if instr == 'EMOS2':
        expr = '#XMMEA_EM'

    if mode == 'IMAGING' or mode == 'imaging':
        xbin = 80
        ybin = 80
        xval = 'X'
        yval = 'Y'

    elif instr == 'EPN' and (mode == 'TIMING' or mode == 'timing'):
        xbin = 1
        ybin = 1
        xval = 'RAWX'
        yval = 'RAWY'
    elif (mode == 'TIMING' or mode == 'timing') and (instr == 'EMOS1' or instr == 'EMOS2'):
        xval = 'RAWX'
        yval = 'TIME'
    else:
        tool_log.error( 'ERROR producing image: {}', image_file)
        return 1

    args = ['table=\"{}\"'.format(event_file), 'expression=\"{}\"'.format(expr), ' withimageset=yes', +\
        'imageset=\"{}\"'.format(image_file), 'imagebinning=binSize', 'ximagebinsize={}'.format(xbin), 'yimagebinsize={}'.format(ybin), 'xcolumn=\"{}\"'.format(xval), 'ycolumn=\"{}\"'.format(yval)]

    t = W('evselect', args)

    t.run()

    tool_log.info('Image: done: {}', image_file)
    
    return 0


def produce_image_4GUI(event_file, image_file, instr, mode):
    '''
    Generates an image for the XMMextractor GUI for an EPN or EMOS1, EMOS2 event list.

    Args:
        event_file: the path to the event file.
        image_file: the name for the image file.
        instr: the instrument (will be normalized)
        mode: the mode

    Output:
        0 if run successfully; 1 otherwise.
    '''

    tool_log.info(' Producing image from {} .... ', event_file)

    if instr == 'EPN':
        expr = '#XMMEA_EP'
    if instr == 'EMOS1':
        expr = '#XMMEA_EM'
    if instr == 'EMOS2':
        expr = '#XMMEA_EM'

    if mode == 'IMAGING' or mode == 'imaging':
        xbin = 80
        ybin = 80
        xval = 'X'
        yval = 'Y'
        extra_params = []
    elif instr == 'EPN' and (mode == 'TIMING' or mode == 'timing'):
        xbin = 1
        ybin = 1
        xval = 'RAWX'
        yval = 'RAWY'
        extra_params = ["imagebinning=binSize", "ximagebinsize={}".format(xbin), "yimagebinsize={}".format(ybin)]
    elif (mode == 'TIMING' or mode == 'timing') and (instr == 'EMOS1' or instr == 'EMOS2'):
        xval = 'RAWX'
        yval = 'TIME'
        extra_params = []
    else:
        tool_log.error( 'ERROR producing image for GUI: {}', image_file)
        return 1

    #image_file = image_file[image_file.rfind('/') + 1:]

    inargs = ['table={}'.format(event_file), 'expression={}'.format(expr),  'withimageset=yes', \
        'imageset={}'.format(image_file), 'xcolumn={}'.format(xval), 'ycolumn={}'.format(yval)] + extra_params

    t = W('evselect', inargs)
    t.run()

    tool_log.info('Image: done')

    return 0


def return_expid(filename):
    """
    Returns the EXPIDSTR keyword in a fashion INXEXPID, being X the mode and
    IN the prefix for the instrument.
    
    Args:
        filename: the file.
    
    Output:
        the EXPIDSTR.
    """

    prefix = ''
    instr = pyutils.get_key_word(filename, 'INSTRUME')

    if instr == 'OM':
        prefix = 'OM'
    if instr == 'EMOS1':
        prefix = 'M1'
    if instr == 'EMOS2':
        prefix = 'M2'
    if instr == 'RGS1':
        prefix = 'R1'
    if instr == 'RGS2':
        prefix = 'R2'
    if instr == 'EPN':
        prefix = 'PN'

    EXPIDSTR = 'unknown'

    try:
        with fits.open(filename) as ev:
            for i in range(0, len(ev)):
                if 'EXPIDSTR' in ev[i].header:
                    EXPIDSTR = ev[i].header['EXPIDSTR']
                    return prefix + EXPIDSTR

            if EXPIDSTR == 'unknown':
                for i in range(0, len(ev)):
                    if 'EXP_ID' in ev[i].header:
                        EXPIDSTR = 'X' + ev[i].header['EXP_ID'][-3:]
                        return prefix + EXPIDSTR
    except FileNotFoundError:
        warnings.warn('Could not open file {}.'.format(filename))
        return None

    return EXPIDSTR


def produce_smooth_image(event_file, image_file, smooth_file, instr, mode):
    '''
    Produces a smooth image.

    Args:
        event_file: the event file.
        immage_file: the image_file.
        smooth_file: name of the output file.
        instr: name of the instrument.
        mode: the mode of the observation.
    
    Output: 
        0 (success) or 1 (error).
    '''

    tool_log.info('   #> Producing smooth image from {} .... ', event_file)

    if instr == 'EPN':
        expr = '#XMMEA_EP'
    if instr == 'EMOS1':
        expr = '#XMMEA_EM'
    if instr == 'EMOS2':
        expr = '#XMMEA_EM'

    if mode == 'IMAGING' or mode == 'imaging':
        xbin = 80
        ybin = 80
        xval = 'X'
        yval = 'Y'

    elif instr == 'EPN' and (mode == 'TIMING' or mode == 'timing'):
        xbin = 1
        ybin = 1
        xval = 'RAWX'
        yval = 'RAWY'
    elif (mode == 'TIMING' or mode == 'timing') and (instr == 'EMOS1' or instr == 'EMOS2'):
        xbin = 1
        ybin = 1
        xval = 'RAWX'
        yval = 'TIME'
    else:
        tool_log.error( 'ERROR producing smooth image.')
        return 1

    inargs = ['table=\"{}\"'.format(event_file), 'expression=\"{}\"'.format(exptr), 'withimageset=yes', + \
        'imageset=\"{}\"'.format(image_file), 'imagebinning=binSize', 'ximagebinsize={}'.format(xbin), 'yimagebinsize={}'.format(ybin), 'xcolumn=\"{}\"'.format(xval), 'ycolumn=\"{}\"'.format(yval)]

    t = W('evselect', inargs)
    t.run()


    cmd = "(asmooth inset=\"{}\" outset=\"{}\" smoothstyle=simple convolverstyle=gaussian width=2.5)".format(image_file, smooth_file)
    if os.system(cmd):
        warnings.warn('Could not run asmooth')
        tool_log.error('Could not run asmooth.')
        return 1

    shutil.copy(smooth_file, obs.image_dir)

    tool_log.info('Smooth image created. Ok')
    return 0


def GTI_file_info(event_file):
    '''
    Reads a GTI event file and returns the ONTIME as a float. Also updates
    the proper GTI log file.
    
    Args:
        event_file: the GTI event file.
    
    Output: 
        GTI_time as string*.

    * To avoid exceptions if the timing is unknown.
    '''

    instr = pyutils.get_key_word(event_file, 'INSTRUME')
    GTI_time = pyutils.get_key_word('ONTIME')

    tool_log.info('#> Information from GTI file: {}', event_file)
    tool_log.info('{} Sum of all Good Time Intervals: {}', instr, GTI_time)

    return GTI_time


def get_OM_science_modes(obs):
    '''
    Returns as a tuple the modes available for a given observation for the OM
    instrument.
    
    Args:
        obs: the observation as an object.
    
    Output: 
        the tuple. 0 indicates that the mode is not available. 1 indicates
    the oposite.

    The order is: omichain, omfchain, omgchain.
    '''

    omichain = False
    omfchain = False
    omgchain = False

    for expo in obs.get_instrument('OM').get_exposures():
        for p in obs.get_instrument('OM').get_exposure(expo).get_products():
            for task in p.get_tasks():                
                if task.task_name == 'omichain':
                    omichain = True
                if task.task_name == 'omfchain':
                    omfchain = True
                if task.task_name == 'omgchain':
                    omgchain = True


    return (omichain, omfchain, omgchain)



def get_region_SP_creation(exposure,t):
    '''
    Returns the source information for a given exposure object.
    
    Args:
        exposure: the exposure as an object from XMMextractor_classes.
    
    Output: 
        returns as a tuple (shape, x, y, L1, L2 and the radius). If one of
    them is 0, then that implies that that value is not used in that geometry.
    '''

    shape = ''
    src_xc = 0
    src_yc = 0
    src_L1 = 0
    src_L2 = 0
    src_r = 0

    region = ''
    prod = exposure.get_product('spectra')
    val = ""
    if t == 'SRC':
        val = 'srcexp'
    else:
        val = 'backexp'
        
    region = ''

    task = prod.get_task('especget','sp_creation')
    expr = task.get_param(val).param_value
    if expr == '':
        return ('no shape', 0, 0, 0 ,0, 0)
    expr = expr.replace(expr[0:expr.index(')') + 2], '')

    shape = ''
    if 'box' in expr:
        shape = 'box'
    if 'annulus' in expr:
        shape = 'annulus'
    if 'circle' in expr:
        shape = 'circle'
    if 'point' in expr:
        shape = 'point'
    if 'line' in expr:
        shape = 'line'

    coords = expr[expr.index('(') + 1:expr.index(')')]
    coords = coords.split(',')
    src_xc = coords[0]
    src_yc = coords[1]

    if shape == 'box' or shape == 'annulus':
        src_L1 = coords[2]
        src_L2 = coords[3]

    if shape == 'circle':
        src_r = coords[2]
    else:
        src_r = 0.0

    return (shape, src_xc, src_yc, src_L1, src_L2, src_r)


def get_region_LC(exposure,t):
    '''
    Returns the source information for a given exposure object.
    
    Args:
        exposure: the exposure as an object from pyOAL.
    
    Output: returns as a tuple (shape, x, y, L1, L2 and the radius). If one of
    them is 0, then that implies that that value is not used in that geometry.
    '''

    shape = ''
    src_xc = 0
    src_yc = 0
    src_L1 = 0
    src_L2 = 0
    src_r = 0

    region = ''
    prod = exposure.get_product('lightcurve')
    exp = ""
    if t == 'SRC':
        exp = 'src_filtering'
    else:
        exp = 'bkg_filtering'
        
    
    task = prod.get_task('evselect','src_filtering')        
    expr = task.get_param('expression').param_value
    if expr == '':
        return ('no shape', 0, 0, 0 ,0, 0)
    expr = expr.replace(expr[0:expr.index(')') + 2], '')

    shape = ''
    if 'box' in expr:
        shape = 'box'
    if 'annulus' in expr:
        shape = 'annulus'
    if 'circle' in expr:
        shape = 'circle'
    if 'point' in expr:
        shape = 'point'
    if 'line' in expr:
        shape = 'line'

    coords = expr[expr.index('(') + 1:expr.index(')')]
    coords = coords.split(',')
    src_xc = coords[0]
    src_yc = coords[1]

    if shape == 'box' or shape == 'annulus':
        src_L1 = coords[2]
        src_L2 = coords[3]

    if shape == 'circle':
        src_r = coords[2]
    else:
        src_r = 0.0

    return (shape, src_xc, src_yc, src_L1, src_L2, src_r)


def build_expression(instr, mode, shape, values, current_log = ''):
    '''
    Writes an expression for the source or the background.
    
    Args:
        instr: the instrument.
        mode: the observation mode (timing or imaging).
        shape: the shape wanted for the expression.
        values: the values wanted.
    
    Output: 
        the expression as a string or empty string if the condiitons were
    not fit.
    '''

    if obsmode.upper() == 'IMAGING':
        xval = 'X'
        yval = 'Y'
    elif obsmode.upper() == 'TIMING' and instr == 'EPN':
        xval = 'RAWX'
        yval = 'RAWY'
    elif obsmode.upper() == 'TIMING' and (instr == 'EMOS1' or instr == 'EMOS2'):
        xval = 'RAWX'
        yval = 'TIME'

    expr = ''

    if shape == 'circle' and len(values) == 3:
        xc = values[0]
        yc = values[1]
        rc = values[2]
        expr = '(({},{}) in {}({},{},{}))'.format(xval, yval, shape, xc, yc, rc)
    elif shape == 'circle' and len(values) == 4:
        xc = values[0]
        yc = values[1]
        if values[2] > values[3]:
            rc = values[2]
        else:
            rc = values[3]
        expr = '(({},{}) in {}({},{},{}))'.format(xval, yval, shape, xc, yc, rc)

    if shape == 'annulus' and len(values) == 4:
        xc = values[0]
        yc = values[1]
        r1 = values[2]
        r2 = values[3]
        expr = '(({},{}) in {}({},{},{},{}))'.format(xval, yval, shape, xc, yc, r1, r2)

    if shape == 'box' and len(values) == 4:
        xc = values[0]
        yc = values[1]
        L1 = values[2]
        L2 = values[3]
        expr = '(({},{}) in {}({},{},{},{}))'.format(xval, yval, shape, xc, yc, L1, L2)

    if expr == '':
        tool_log.warning('\n Something wrong happened while writting an expression.')
    if current_log != '':
        tool_log.info('Expression for the given parameters: {}', expr)

    return expr


def prepare_region_log_file(instr, exposure, current_product, current_log, src_info = '', \
    bkg_info = ''):
    '''
    Writes into a log file regarding a region.
    
    Args:
        instr: the instrument.
        exposure:: an exposure object.
        current_product: the type of the product as string.
        current_log,: a log file.
        src_info (optional): info regarding region if the one read from the expression
    is not good (src).
        bkg_info (optional): info regarding region if the one read from the expression
    is not good (bkg).
    '''

    if src_info == '' and bkg_info == '':
        src_shape, src_xc, src_yc, src_L1, src_L2, src_R = get_region(exposure,'SRC')
        bkg_shape, bkg_xc, bkg_yc, bkg_L1, bkg_L2, bkg_R = get_region(exposure,'BKG')
    elif src_info != '' and bkg_info == '':
        bkg_shape, bkg_xc, bkg_yc, bkg_L1, bkg_L2, bkg_R = get_region(exposure,'BKG')
        src_shape, src_xc, src_yc, src_L1, src_L2 = src_info
    elif src_info == '' and bkg_info != '':
        src_shape, src_xc, src_yc, src_L1, src_L2, src_R = get_region(exposure,'SRC')
        bkg_shape, bkg_xc, bkg_yc, bkg_L1, bkg_L2 = bkg_info
    elif src_info != '' and bkg_info != '':
        bkg_shape, bkg_xc, bkg_yc, bkg_L1, bkg_L2 = bkg_info
        src_shape, src_xc, src_yc, src_L1, src_L2 = src_info

    if current_product == 'LC':
        product = 'lightcurve'
    else:
        product = 'spectra'

    tool_log.debug("   #> {} processing...", product)
    tool_log.debug("   #> Centroid of source to analyze (Physical Units)")
    tool_log.debug("   #> INSTRUMENT: {} EXPOSURE: {}", instr, exposure.expid)
    tool_log.debug("         SRC Region : {}", src_shape)
    tool_log.debug("         SRC XC     : {}", src_xc)
    tool_log.debug("         SRC YC     : {}", src_yc)

    if src_shape == "circle":
        if src_L1 == 0:
            if src_L2 != 0:
                src_r = src_L2
        else:
            if src_L2 == 0:
                src_r = src_L1
        tool_log.info('         SRC R    : {}', src_r)
    if src_shape == "annulus":
        tool_log.info('         SRC R1 - SRC R2    : {} - {}', src_L1, src_L2)
        if src_L1 > src_L2:
            tool_log.warning(' source inner radius is bigger than outer radius')
    if src_shape == "box":
        tool_log.info("         SRC BOX    : {} {}", src_L1, src_L2)

    # BKG
    tool_log.debug("         BKG Region : {}", bkg_shape)
    tool_log.info("         BKG XC     : {}", bkg_xc)
    tool_log.info("         BKG YC     : {}", bkg_yc)
    
    if bkg_shape == "circle":
        if bkg_L1 == 0:
            if bkg_L2 != 0:
                bkg_r = bkg_L2
        else:
            if bkg_L2 == 0:
                bkg_r = bkg_L1
        tool_log.info('         BKG R    : {}', bkg_r)
    if bkg_shape == "annulus":
        tool_log.info('         BKG R1 - BKG R2    : {} - {}', bkg_L1, bkg_L2)
        if bkg_L1 > bkg_L2:
            tool_log.warning('WARNING: bkg inner radius is bigger than outer radius')
    if bkg_shape == "box":
        tool_log.info("         BKG BOX    : {} {}", bkg_L1, bkg_L2)
        area_factor = exposure.get_products(product)[0].get_tasks("xmmextractor","details")[0].get_params("areafactor")[0].param['default']
        tool_log.info("         BKG AREA SCALE FACTOR : {} (Only area taken into account)", area_factor)
        if bkg_L1 == 6000 or bkg_L2 == 6000 or src_L1 == 6000 or src_L2 == 6000:
            tool_log.warning("         WARNING : PN regions : Maximum allowed radius reached")

    tool_log.info('Finished writing region log.')
    return 0


def event_file_info(event_file, instr):
    '''
    Gathers the information of a given event file.

    Args:
        event_file: the event file.
        instr: the instrument.

    Output: 
        tuple with (dmode, filter, submode, utc_obsdate, obs_duration, ontime, livetime, t_start, t_end).
    '''

    ext = 1
    if instr == 'OM':
        ext = 0

    dmode = pyutils.get_key_word(event_file, 'DATAMODE', ext)
    submode = pyutils.get_key_word(event_file, 'SUBMODE', ext)
    filter_info = pyutils.get_key_word(event_file, 'FILTER', ext)
    utc_obsdate = pyutils.get_key_word(event_file, 'DATE-OBS', ext)
    obs_duration = pyutils.get_key_word(event_file, 'DURATION', ext)
    ontime = pyutils.get_key_word(event_file, 'ONTIME', ext)
    livetime = pyutils.get_key_word(event_file, 'LIVETIME', ext)
    exposure = return_expid(event_file)
    t_start = pyutils.get_key_word(event_file, 'TSTART', ext)
    t_end = pyutils.get_key_word(event_file, 'TSTOP', ext)

    # if tstart or tend are not defined, the values will be taken from the
    # data. Same for duration.
    if t_start =='unknown' or t_end == 'unknown':
        with fits.open(event_file) as ev:
            data_time = ev[ext].data['TIME']
            t_start = data_time[0]
            t_end = data_time[-1]

    if obs_duration == 'unknown':
        obs_duration = t_end - t_start

    # Writting to the given log...

    tool_log.info('    #> Information from Event File: {}', event_file)
    tool_log.info("       {} OBS DATE          : {}", instr, utc_obsdate)
    tool_log.info("       {} OBS STARTING TIME : {} (1st Event)", instr, t_start)
    tool_log.info("       {} OBS ENDING TIME   : {} (Last Event)", instr, t_end)
    tool_log.info("       {} EXPOSURE ID  : {} ", instr, exposure)
    tool_log.info("       {} OBS DURATION : {}   ", instr, obs_duration)
    tool_log.info("       {} OBS ONTIME   : {}", instr, ontime)
    tool_log.info("       {} OBS LIVETIME : {}", instr, livetime)
    tool_log.info("       {} D MODE       : {}  ", instr, dmode)
    tool_log.info("       {} SUBMODE      : {}   ", instr, submode)
    tool_log.info("       {} FILTER       : {} ", instr, filter_info)

    return (dmode, filter_info, submode, utc_obsdate, obs_duration, ontime, livetime, t_start, t_end)


def postcard_from_XSA(obsid, filename = None):
    '''
    Downloads an images given the observation id using the postcard format from astroquery.

    Args:
        obsid: the observation id.
        filename: the name of the image. None by default - will use the name from XSA.

    Ouput:
        img_name: the name of the downloaded image.
        1 if run into errors.
    '''

    try:
        img_name = XMMNewton.get_postcard(str(obsid), filename = None)
    except:
        return 1

    return img_name


def unpackODF(obs_file):
    '''
    Unpacks the OBS file tar.gz file.

    Args:
        obs_file: the path to the .tar.gz file.

    Output:
        0 when finished, 1 if any errors were found.
    '''

    try:
        targz_file = tarfile.open(obs_file, 'r:gz')
    except FileNotFoundError:
        return 1

    targz_file.extractall()
    targz_file.close()
    os.remove(os.path.basename(obs_file))


    tar_file = glob.glob('*.TAR')

    if len(tar_file) != 0:
        tar_file = tar_file[0]
        tar = tarfile.open(tar_file, 'r:')
        tar.extractall()
        tar.close()
        os.remove(os.path.basename(tar_file))

    return 0


def check_for_connection():
    '''
    Checks if there's internet connection available.

    Args:
        none

    Output: bool.
    '''

    host = 'http://google.com'

    try:
        urllib.request.urlopen(host)
        return True
    except:
        return False


def xsaRequest_image(input_obs):
    '''
    Downloads a PPS-level image from the XSA using astroquery.

    Args:
        input_obs: the OBSID for the download.

    Output:
        0 when finished, 1 if an exception is found.
    '''

    fileName = input_obs+'.FTZ'

    try:
        XMMNewton.download_data(input_obs, filename=fileName,level = 'PPS', extension='FTZ',name='OIMAGE', verbose = True)
    except:
        return 1

    return 0


def download_ODF_from_XSA(input_obs, level, username = '', password = ''):
    '''
    Download the product data for an input file.

    Args:
        input_obs: the obsid.
        level: either 'ODF', 'PPS'.

    Output:
        0 when finished, 1 if any errors occurred.
    '''
    
    try:
        XMMNewton.download_data(str(input_obs), level = level)
    except:
        return 1
    
    return 0



# version.py 
#
# ESA (C) 2000-2023
#
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#
# 16-Apr-2023

VERSION = '1.2.1'


# ESA (C) 2000-2021
#
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

# omdrifthist.py

from .version import VERSION
from pysas import SAS_RELEASE, SAS_AKA

__version__ = f'omdrifthist (omdrifthist-{VERSION}) [{SAS_RELEASE}-{SAS_AKA}]'


import numpy as np
from matplotlib.patches import ConnectionPatch
import matplotlib.pyplot as plt
from matplotlib.ticker import PercentFormatter
import warnings
import sys
import re
import os
import glob
import math
import pysas.pyutils.pyutils as pyutils
from astropy.io import fits
from astropy.table import Table
import time
import pysas.pysasplot_utils.pysasplot_utils as sasplt
from logger import get_logger

logger = get_logger('omdrifthist')

########################################
############ CALC FUNCTIONS ############
########################################

def euclid_modulus(point):
    """
    Returns the modulus of two given points.

    Args:
        point: the coordinates of a point in a 2D space.

    Output:
        modulus: the calculated modulus.
    """

    a, b = point
    modulus = np.sqrt(a**2 + b**2)

    return modulus


def incremental_drift(history, x0, y0, rpixel):
    """
    Calculates the incremental drift roll and the incremental drift coordinates.

    Args:
        history: the 2D matrix containing the x data and y data in separate
    columns.
        x0: original x coordinate from the file.
        y0: original y coordinate from the file.
        rpixel: the radius for the pixel adapted from trackradius.

    Output:
        incdrift: the list containing the incremental drift size.
        increments: the array of increments in the drift roll
        n_out: the number of pixels outside the original radius.
        n_out_0: the number of pixels outside the original radius at x0, y0
    """

    logger.debug('Calculating incremental drift arrays...')

    n_out = 0
    incdrift = [euclid_modulus(history[0])]
    increments = [history[0]]
    n_out_0 = 0

    for i in range(1, len(history)):
        incdrift.append(euclid_modulus(history[i] - history[i - 1]))
        increments.append(history[i] - history[i - 1])
        r0 = np.sqrt((x0 - history[i][0]) ** 2 + (y0 - history[i][1]) ** 2)
        r = euclid_modulus(history[i])
        if r > rpixel:
            n_out = n_out + 1
        if r0 > rpixel:
            n_out_0 = n_out_0 + 1

    increments = np.vstack(increments)
    logger.debug('Incremental drift arrays calculated.')

    return incdrift, increments, n_out, n_out_0


def get_more_recent_ccf(abspaths):
    """
    Returns the absolute path of the most recent CCF file.

    Args:
        abspaths: the absolute path of the files to evaluate.

    Output:
        the absolute path of the most recent file.
    """

    maxn = 0
    for f in abspaths:
        f_aux = os.path.basename(f)
        num = int(re.search(r'\d+', f_aux).group(0))
        if num > maxn:
            maxn = num
            maxf = f
            
    return maxf


def get_platescale(filt):
    """
    Searches in the optical monitor calibration file for the value of the plate
    scale for the given filter.

    Args:
        filt: the filter used in the observation.

    Output:
        platescale: the value for the plate scale as float.
    """

    logger.debug('Looking for platescale...')

    try:
        ccfpath = os.environ['SAS_CCFPATH']
    except KeyError:
        logger.error('Could not locate variable SAS_CCFPATH. Quitting...')
        sys.exit(0)


    folders = ccfpath.split(':')
    if len(folders) == 1:
        om_ccf_file = glob.glob(os.environ['SAS_CCFPATH'] + '/OM_ASTR*')
        om_ccf_file = get_more_recent_ccf(om_ccf_file)
        logger.debug('CCF file used: {}.', om_ccf_file)
        search_extension = 'FILTER-' + filt.upper()
        platescale = pyutils.get_key_word(om_ccf_file, 'PLTSCALE', search_extension)
    else:
        om_ccf_list = []
        logger.debug('List of folders found in CCF path: {}', folders)
        for folder in folders:
            try:
                om_ccf_file = os.path.abspath(glob.glob(folder + '/OM_ASTR*')[0])
            except IndexError:
                logger.debug('{} does not contain OM astronometry calibration file.', folder)
                continue
            om_ccf_list.append(om_ccf_file)
        logger.debug('Found the following OM CCF compatible files: {}', om_ccf_list)
        latest_ccf = get_more_recent_ccf(om_ccf_list)
        logger.debug('CCF file used: {}.', latest_ccf)
        search_extension = 'FILTER-' + filt.upper()
        platescale = pyutils.get_key_word(latest_ccf, 'PLTSCALE', search_extension)


    if platescale == 'unknown':
        logger.warning('Could not find the platescale keyword in the CCF. Will use the most frequen value.')
        platescale = 0.476513
    else:
        logger.info('Platescale retrieved from ccf file. Value: {}.', platescale)
    return float(platescale)


def evaluate_zerodrift(tracking_header):
    """
    Checks if the header contains the ZERODRIFT keyword, which implies that
    the file is not suitable for working with omdrifthist.

    Args:
        tracking_header: a copy of the header for the tracking history file.

    Output:
        1 when finished; will exit if ZERODRIFT is found.
    """

    zerodrift_flag = True

    try:
        zerodrift = tracking_header['ZERODRIFT']
    except KeyError:
        zerodrift_flag = False

    if zerodrift_flag:
        logger.error('The file was marked as inadequate to work with omdrifthist.')
        sys.exit(0)

    return 1


########################################
########## PLOTTING FUNCTIONS ##########
########################################

def omdrifthist1(datax, datay, x0, y0, r0, scale = '', nbins = 50, added_text ='', output = 'plot', fig_title = 'OM tracking plot'):
    """
    Creates a tracking plot form the input points, with count histograms for the
    x - y axis.This task provides graphical and statistical information on the OM tracking history.

    Args:
        datax: the array of x-coordinates.
        datay: the array of y-coordinates.
        x0: the x coordinate for the center of the circle.
        y0: the y coordinate for the center of the circle.
        r0: the radius of the circle.
        nbins: number of bins. 50 by default.
        scale: the y scale.
        added_text: text to be added.
        output: the name of the final file. plot1.png by default
        fig_title: the name of the main figure..

    Output:
        1 when finished, 0 if could not finish the plot.
    """

    logger.info('Working on plot 1.')
    # evaluating data

    if len(datax) != len(datay):
        logger.warning('The coordinate length is not the same for bot axes. Plot 1 will not be produced...')
        return 0

    ####################################
    # definitions for the axes
    left, width = 0.1, 0.55
    bottom, height = 0.1, 0.55
    spacing = 0.005

    coordsA = "data"
    coordsB = "data"

    rect_scatter = [left, bottom, width, height]
    rect_histx = [left, bottom + height + spacing, width, 0.2]
    rect_histy = [left + width + spacing, bottom, 0.2, height]

    # start with a rectangular Figure
    f = plt.figure(figsize=(11, 11))

    ax_scatter = plt.axes(rect_scatter)
    ax_scatter.tick_params(direction='in', top=True, right=True)
    ax_histx = plt.axes(rect_histx)
    ax_histx.set_ylabel('Freq.')
    ax_histx.tick_params(direction='in', labelbottom=False)
    ax_histy = plt.axes(rect_histy)
    ax_histy.tick_params(direction='in', labelleft=False)
    ax_histy.set_xlabel('Freq.')
    # the scatter plot:
    t = np.arange(len(datax))
    ax_scatter.scatter(datax, datay, marker = 'o', label = 'Scattered points', c=t, cmap=plt.get_cmap('summer'))
    
    #drawing arrows...
    for i in range(0, len(datay) - 1):
        if i == 0:
            minx = datax[i]
            miny = datay[i]
            maxx = datax[i]
            maxy = datay[i]
        xyA = (datax[i], datay[i])
        xyB = (datax[i + 1], datay[i + 1])
        r = np.sqrt((x0 - datax[i]) ** 2 + (y0 - datay[i]) ** 2)
        if r < r0:
            con = ConnectionPatch(xyA, xyB, coordsA, coordsB,
                    arrowstyle = "-|>", shrinkA = 5, shrinkB = 5,
                      mutation_scale = 20, fc = "w")
        else:
            con = ConnectionPatch(xyA, xyB, coordsA, coordsB,
                    arrowstyle = "->", shrinkA = 5, shrinkB = 5,
                        mutation_scale = 20, fc = "w")
        ax_scatter.add_artist(con)
        
        if datax[i + 1] > maxx:
            maxx = datax[i + 1]
        if datax[i + 1] < minx:
            minx = datax[i + 1]
        if datay[i + 1] < miny:
            miny = datay[i + 1]
        if datay[i + 1] > maxy:
            maxy = datay[i + 1]

    #### more testing code
    if (r0 + x0) > maxx:
        maxx = r0 + x0
    if (x0 - r0) < minx:
        minx = x0 - r0
    if (r0 + y0) > maxy:
        maxy = r0 + y0
    if (y0 - r0) < miny:
        miny = y0 - r0

    if maxx > maxy:
        absmax = maxx
    else:
        absmax = maxy
    if minx > miny:
        absmin = miny
    else:
        absmin = minx

    binwidth = 0.25
    lim = np.ceil(np.abs([datax, datay]).max() / binwidth) * binwidth
   
    ax_scatter.set_xlim((absmin, absmax))
    ax_scatter.set_ylim((absmin, absmax))

    bins = np.arange(-lim, lim + binwidth, binwidth)
    ax_histx.hist(datax, bins = nbins, color = 'forestgreen', alpha = 0.2, label='x counts')
    ax_histy.hist(datay, bins = nbins, orientation = u'horizontal', color = 'forestgreen', alpha = 0.2,label='y counts')

    ax_histx.set_xlim(ax_scatter.get_xlim())
    ax_histy.set_ylim(ax_scatter.get_ylim())

    circle_region = plt.Circle((x0, y0), r0, color = 'b', fill=False, label = 'Nominal tracking circle')
    ax_scatter.add_patch(circle_region)

    ax_scatter.set_xlabel('x (pixels)')
    ax_scatter.set_ylabel('y (pixels)')

    #ax_histy.text(0.5, 1.5, added_text, fontsize = 9)
    plt.text(1.01, 0.6, added_text, ha='left', va = 'top', transform = ax_histx.transAxes)

    if scale != '':
        try:
            plt.yscale(scale)
        except ValueError:
            warnings.warn('The given scale is not supported. Using linear scale instead.')
            plt.yscale('linear')

    f.suptitle(fig_title, fontsize=19)

    if '2' in pages:
        if out_format.upper() == 'PDF':
            out1 = os.path.splitext(output)[0] + '_1{}.{}'.format('DRIFT_TEMP', out_format)
        else:
            out1 = os.path.splitext(output)[0] + '_{}.{}'.format('1', out_format)
    else:
        out1 = output + '.' + out_format

    logger.debug('Saving plot 1: {}.{} ({})', output, out_format, out1)

    try:
        plt.savefig(out1)
    except FileNotFoundError:
        try:
            dirs, rfile = os.path.split(out1)
            os.makedirs(dirs)
            plt.savefig(out1)
        except:
            logger.error('Could not create {}. Could not resolve the given path.', out1)

    if pyutils.is_notebook():
        plt.show()
    
    logger.info('Plot produced: {}', out1)
    
    return 1


def omdrifthist2(roll, nframes, datax, datay, incremental_drift, scale = '', nbins = 50, added_text = '', output = 'plot', fig_title = 'OM tracking'):
    """
    Presents plots with data related to the drift for the tracking histogram:
     - the histogram with the incremental drift sizes.
     - the roll drift according to each frame.
     - the incremental drift spots coordinates.

    Args:
        roll: the roll frames data.
        nframes: an array with the frames for the roll drift plot.
        datax: an array with the x coordinates of the incremental drift spots.
        datay: an array with the y coordinates of the incremental drift spots.
        incremental_drift: the incremental drift array.
        scale: the wanted scale for the y axis.
        nbins: the number of bins for the histogram.
        added_text: extra text to be added to the figure.
        output: the output file.
        fig_title: name of the main figure.

    Output:
        1 when finished.
    """

    logger.info('Working on plot 2.')

    roll = np.array(roll) * 180 / math.pi
    f, ax = plt.subplots(2, 2, figsize = (8, 10), sharey=False, sharex=False)

    ax1 = ax[1,0]
    ax[1,0].plot(nframes, roll, marker = '.')
    high, bins, _ = ax[0,1].hist(incremental_drift, color = 'forestgreen', fill = True, bins = nbins, alpha = 0.2, density = False)

    # plot titles...
    ax[1,1].title.set_text('Incremental drift spot diagram')
    ax[1,0].title.set_text('Roll drift plot')
    ax[0,1].title.set_text('Drift size histogram')

    # left empty to add text
    ax[0,0].axis('off')
    ax[0,0].text(-0.1,0.3, added_text, fontsize=10)

    t = np.arange(len(datax))
    cs = ax[1,1].scatter(datax, datay, marker = 'o', c=t, cmap=plt.get_cmap('summer'), label = 'Scattered points')
    cbar = f.colorbar(cs, orientation = 'vertical')
    cbar.ax.set_ylabel('Frame number')

    ax[0,1].set_xlabel('Incremental drift size (pixels)')
    ax[0,1].set_ylabel('Counts')
    # lowering histogram plot:
    ax[0,1]
    ax[1,0].set_xlabel('Frame number')
    ax[1,0].set_ylabel('Roll drift (degrees)')
    ax[1,1].set_xlabel('x (pixels)')
    ax[1,1].set_ylabel('y (pixels)')
    skip_pos = fig_title.find(':') + 1
    fig_title = fig_title[:skip_pos] + "\n" + fig_title[skip_pos:]
    f.suptitle(fig_title, x = 0.3, fontsize = 16)
    f.tight_layout()
    
    # to add in future?
    if scale != '':
        try:
            plt.yscale(scale)
        except ValueError:
            warnings.warn('The given scale is not supported. Using linear scale instead.')
            plt.yscale('linear')

    if '1' in pages:
        if out_format.upper() == 'PDF':
            out2 = os.path.splitext(output)[0] + '_2{}.{}'.format('DRIFT_TEMP', out_format)
        else:
            out2 = os.path.splitext(output)[0] + '_{}.{}'.format('2', out_format)
    else:
        out2 = output + '.' + out_format

    logger.debug('Saving plot 2: {}.{} ({})', output, out_format, out2)

    try:
        plt.savefig(out2)
    except FileNotFoundError:
        try:
            dirs, rfile = os.path.split(out2)
            os.makedirs(dirs)
            plt.savefig(out2)
        except:
            logger.error('Could not create {}. Could not resolve the path.', out2)

    if pyutils.is_notebook():
        plt.show()
    logger.info('Plot produced: {}', out2)

    return 1


########################################


def run(iparsdic):
    logger.warning(f'Executing {__file__} {iparsdic}')

    t_start = time.time()
    fits_file = iparsdic['set']
    trackradius = float(iparsdic['trackradius'])
    global pages
    pages = str(iparsdic['pages'])
    output = str(iparsdic['plotfile'])
    plot_abs_path = os.path.abspath(output)
    output_base = os.path.basename(plot_abs_path)
    output_base = os.path.splitext(plot_abs_path) # splits the format and the plotfile
    
    global out_format
    
    if output_base[1] == '':
        output = plot_abs_path
        out_format = 'pdf'
        logger.warning('Format not found in the plotfile parameter. Using {}.{} as default.', plot_abs_path, out_format)
    else:
        output_base, out_format = output_base
        out_format = out_format.replace('.', '')
        output = output_base
        logger.debug('Using {}.{}', output, out_format)

    logger.debug('Working with {}.{}', output, out_format)
    nbins = int(iparsdic['nbins'])
    verbosity = int(os.environ['SAS_VERBOSITY'])

    if not out_format.upper() in ('PNG', 'PDF'):
        logger.error('Format {} not understood. Please use only pdf or png formats.', out_format)
        sys.exit(0)

    TBD = 10

    if not os.path.isfile(fits_file):
        logger.error('The input file does not exist.')
        sys.exit(0)

    datamode = pyutils.get_key_word(fits_file, 'DATAMODE')
    if datamode != 'TRACKING':
        logger.error('Invalid datamode ({}) for tracking history.', datamode)
        sys.exit(0)

    try:
        binbpe = pyutils.get_key_word(fits_file, 'BINBPE')
        exp_id = pyutils.get_key_word(fits_file, 'EXP_ID')
        obj = pyutils.get_key_word(fits_file, 'OBJECT')
        obs_id = pyutils.get_key_word(fits_file, 'OBS_ID')
        filt = pyutils.get_key_word(fits_file, 'FILTER')
        frametime = pyutils.get_key_word(fits_file, 'FRMTIME')
    except FileNotFoundError:
        logger.error('Could not open the fits file. Quitting...')
        sys.exit(0)

    platescale = get_platescale(filt)
    trackradius = trackradius / platescale

    if os.path.isfile(fits_file):
        with fits.open(fits_file) as f:
            prim_h = f[0].header
            tracking_header = f[1].header
            t = Table(f[1].data)
    else:
        logger.error('Could not find the fits file. Quitting...')
        sys.exit(0)

    evaluate_zerodrift(tracking_header)

    logger.debug('File {} read.', fits_file)
    rpixel = trackradius

    try:
        nframes = t['FRAME']
        dx = t['DX']
        dy = t['DY']
        t_roll = t['ROLL']
        quality = t['QUALITY']
        nggs = t['NGGS']
    except KeyError:
        pyutils.pymodhdu(fits_file, 1, 'ZERODRIFT', 1, comment = 'File not suitable for working with omdrifthist.')
        logger.error('At least one column could not be loaded.')
        sys.exit(0)

    if len(nframes) == 0:
        pyutils.pymodhdu(fits_file, 1, 'ZERODRIFT', 1, comment = 'File not suitable for working with omdrifthist.')
        logger.warning('Not a valid tracking history file. Will write extension to notify.')
        sys.exit(0)

    nrows = len(t)

    # create an array with the calculated quality
    quality_array = quality / 1000000
    quality_array = np.sqrt(quality_array / (2 * nggs - 1))

    # verify the data (bad tracking files)
    nbad = 0
    nhist = 0

    x = []
    y = []
    roll = []

    for i in range(0, nrows):
        if quality_array[i] < TBD and nggs[i] > 3:
            nhist = nhist + 1
            x.append(dx[i] / 1000)
            y.append(dy[i] / 1000)
            roll.append(math.asin(t_roll[i] / 1000000))
        else:
            nbad = nbad + 1

    x0 = x[0]
    y0 = y[0]

    # creates the history of appearance of the filtered x,y points
    history = np.array([x, y])
    history = np.matrix.transpose(history)

    incdrift, increments, nout, n_out_0 = incremental_drift(history, x0, y0, rpixel)

    inc_x = increments[:,0]
    inc_y = increments[:,1]
    
    meandrift = sum(incdrift) / (len(nframes) - 1)
    del(increments)

    fig_title = 'OM tracking history plot: object {0}.'.format(obj)
    if '1' in str(pages):
        text = "Observation ID: {0}. \nExposure ID: {1}.\nFilter: {2}. Plate-scale: {3} \narcsecs/pixel.\n".format(obs_id, exp_id, filt, platescale)
        text = text + "Tracking frame time: {0} s.\nNumber of good tracking frames: {1}.\nTracking radius: {2}.\n".format(round(float(frametime)/1024, 2), nhist, round(trackradius))
        text = text + "Number of bad tracking frames: {0}.\n".format(nbad)
        omdrifthist1(x, y, x0, y0, rpixel, scale = '', nbins = nbins, added_text = text, output = output, fig_title = fig_title)
    
    if '2' in str(pages):
        text = "Observation ID: {0}. \nExposure ID: {1}.\nFilter: {2}. Plate-scale: {3} arcsecs/pixel.\n".format(obs_id, exp_id, filt, platescale)
        text = text + "Tracking frame time: {0} s.\nNumber of good tracking frames: {1}.\n".format(round(float(frametime)/1024, 2), nhist)
        text = text + "Tracking radius: {:.2f}.\n".format(trackradius)
        text = text + "Number of bad tracking frames: {0}.\n\n".format(nbad)
        text = text + "Percentage of pixels outside radius: {:.2f}%.\n\n".format(nout/len(nframes) * 100)
        #text = text + "Percentage of excursion outside initial\ncircle coordinates: {} %.\n\n".format(n_out_0/len(nframes) * 100)
        text = text + "Mean drift per tracking frame (pixels): {:.2f}.".format(meandrift)
        omdrifthist2(roll, nframes, inc_x, inc_y, incdrift, nbins = nbins, added_text = text, output = output, fig_title = fig_title)

    if '1' in str(pages) and '2' in str(pages) and out_format.upper() == 'PDF':
        logger.debug('More than one PDF created. Merging...')
        pdf_files_found = glob.glob(os.path.splitext(output)[0] + '*DRIFT_TEMP*')
        pdf_files_found.sort()
        logger.debug('sorted list: {}', pdf_files_found)
        output = output.replace('DRIFT_TEMP', '')
        status = sasplt.merge_pdf(pdf_files_found, os.path.splitext(output)[0] + '.{}'.format(out_format))

        if status == 0:
            logger.warning('Could not merge the two PDFs')
        else:
            logger.debug('PDF merged. Cleaning up...')
            for pdffile in pdf_files_found:
                if pdffile == output + '.{}'.format(out_format):
                    pass
                else:
                    os.remove(pdffile)

    t_stop = time.time()
    logger.info('Finished running omdrifthist in {:.2f} seconds.'.format(t_stop - t_start))
    print('Finished running omdrifthist in {:.2f} seconds.'.format(t_stop - t_start))




# odfcontrol.py
#
# Written by: Ryan Tanner
# email: ryan.tanner@nasa.gov
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS. If not, see <http://www.gnu.org/licenses/>.

# odfcontrol.py

"""
odfcontrol.py

"""

# Standard library imports
import os, sys, shutil, glob, numbers
from warnings import warn

# Third party imports

# Local application imports
from pysas import sas_cfg
from ..init_sas import initializesas
from ..wrapper import Wrapper as w
from ..sasutils import generate_logger
from ..sasutils import download_data as dl_data

__all__ = ['ODFobject']


class ODFobject(object):
    """
    Class for observation data files (ODF).

        An odfid (Obs ID) is required.

        data_dir is the base directory where you store all XMM data.

        Data is organized as:
            data_dir = /path/to/data/
            obs_dir = /path/to/data/odfid/
        With subdirectories and files:
                odf_dir  = /path/to/data/odfid/ODF/
                work_dir = /path/to/data/odfid/work/
                SAS_CCF  = work_dir/ccf.cif
                SAS_ODF  = work_dir/*SUM.SAS

                
        Directory links and OdfID information are stored as
        variables in the ODFobject.

            self.odfid    : The Obs ID required to make the ODFobject
            self.data_dir : Path to the data directory
            self.files    : A dictionary containing links to event lists
                            and other important files

                            Default dictionary keys are:
                                'sas_ccf'
                                'sas_odf'
                                'PNevt_list'
                                'M1evt_list'
                                'M2evt_list'
                                'R1evt_list'
                                'R2evt_list'
                                'OMevt_list'

            self.obs_dir  : Path to the observation directory
            self.odf_dir  : Path to the ODF directory
            self.pps_dir  : Path to the PPS directory
            self.work_dir : Path to the working directory

    """

    def __init__(self,odfid,data_dir=None):
        warn(
             """
             The ODFobject class has been depricated. Use ObsID instead.
             ex: 
                 my_obs = pysas.obsid.ObsID(obsid)
             instead of
                 odf = pysas.odfcontrol.ODFobject(obsid)
             """)
        if isinstance(odfid, numbers.Number):
            odfid = str(odfid)
        self.odfid = odfid
        self.data_dir = data_dir
        self.files = {}
        self.__set_obsid()

    def basic_setup(self,data_dir=None,level='ODF',repo='esa',
                    overwrite=False,recalibrate=False,rerun=False,
                    run_epproc=True,run_emproc=True,run_rgsproc=True,
                    run_epchain=False,run_emchain=False,
                    **kwargs):
        """
        Function to do all basic analysis tasks. The function will:

            1. Download data by calling 'download_data'
            2. Call the function 'calibrate_odf'
                A. Run 'cifbuild'
                B. Run 'odfingest'
            2. Run 'epproc' -OR- 'epchain'
            3. Run 'emproc' -OR- 'emchain'
            4. Run 'rgsproc'

        If 'run_epchain' is set to 'True', then 'epproc' will not run.
        If 'run_emchain' is set to 'True', then 'emproc' will not run.

        Inputs:

            data_dir:    Data directory.
            level:       Level of data products ('ODF','PPS','ALL').
            repo:        Download repository ('esa','heasarc','sciserver').
            overwrite:   Remove previous data files and download again.
            recalibrate: Rerun 'cifbuild' and 'odfingest'.
            rerun:       Rerun the *procs or *chains.

        All input arguments for 'download_data' and 'calibrate_odf'
        can be passed to 'basic_setup'.

        'download_data' inputs (with defaults):

            repo            = 'esa'
            level           = 'ODF'
            data_dir        = None
            overwrite       = False
            logger          = None
            proprietary     = False
            credentials_file= None
            encryption_key  = None
            PPS_subset      = False
            instname        = None
            expflag         = None
            expno           = None
            product_type    = None
            datasubsetno    = None
            sourceno        = None
            extension       = None
            **kwargs

        'calibrate_odf' inputs (with defaults):
               
            obs_dir        = None
            sas_ccf        = None
            sas_odf        = None
            cifbuild_opts  = []
            odfingest_opts = []
            recalibrate    = False
            logger         = None

        Input arguments for 'epproc', 'emproc', and 'rgsproc' can also be 
        passed in using 'epproc_args', 'emproc_args', or 'rgsproc_args' 
        respectively (or 'epchain_args' and 'emchain_args'). By defaut 
        'epproc', 'emproc', and 'rgsproc' will not rerun if output files 
        are found, but they can be forced to rerun by setting 'rerun=True' 
        as an input to 'basic_setup'.

        Examples for use:

            odf.basic_setup()

                - Uses the defaults.

            odf.basic_setup(repo='heasarc')

                - Uses the defaults, but downloads data from the HEASARC.

            odf.basic_setup(overwrite=True)

                - Will erase any previous data files for the Obs ID and 
                  download a fresh set of data files.

            odf.basic_setup(recalibrate=True)

                - Will rerun cifbuild and odfingest to generate new 
                  ccf.cif and *SUM.SAS files.

            odf.basic_setup(rerun=True)

                - Will **not** download new files, but will rerun 'epproc',
                  'emproc', and 'rgsproc' and create new event lists.

            odf.basic_setup(repo='heasarc',
                            epproc_args=['withoutoftime=yes'])

                - Downloads data from the HEASARC and runs 'epproc' with the
                  'withoutoftime' option.

            odf.basic_setup(run_epchain=True,
                            run_emchain=True)

                - Will run 'epchain' and 'emchain' instead of 'epproc' and
                  'emproc'.

            odf.basic_setup(run_epproc=False,
                            run_emproc=False)

                - Will not run 'epproc' or 'emproc'. Will only run 'rgsproc'
                  by default.

            odf.basic_setup(run_epproc=False,
                            run_emproc=True,
                            run_rgsproc=False)

                - Will only run 'emproc', **not** 'epproc' or 'rgsproc'.

            odf.basic_setup(repo='heasarc',encryption_key='XXXXXXXXXXXXXXX')

                - Uses the defaults, but downloads *proprietary* data from 
                  the HEASARC. Must provide an encryption key, an alpha-numeric
                  string with 30 characters.

            odf.basic_setup(proprietary=True)

                - Uses the defaults, but downloads *proprietary* data from 
                  the XSA at ESA. Astroquery will ask for user's Cosmos
                  username and password.

        """

        # Where are we?
        startdir = os.getcwd()

        # Brief check to see if data_dir was 
        # given on odfobject creation.
        if self.data_dir != None:
            data_dir = self.data_dir

        # Start checking data_dir
        if data_dir == None:
            if sas_cfg.config.has_option('sas','data_dir'):
                data_dir = sas_cfg.get_setting('data_dir')
            else:
                data_dir = '/does/not/exist'
            if os.path.exists(data_dir):
                self.data_dir = data_dir
            else:
                self.data_dir = startdir
        else:
            self.data_dir = data_dir

        # If data_dir was not given as an absolute path, it is interpreted
        # as a subdirectory of startdir.
        if self.data_dir[0] != '/':
            self.data_dir = os.path.join(startdir, self.data_dir)
        elif self.data_dir[:2] == './':
            self.data_dir = os.path.join(startdir, self.data_dir[2:])

        # Check if data_dir exists. If not then create it.
        # Save comments for the logger created later.
        logcomment = ''
        if not os.path.isdir(self.data_dir):
            logcomment = f'{self.data_dir} does not exist. Creating it!'
            os.mkdir(self.data_dir)
            logcomment = logcomment + '\n' + f'{self.data_dir} has been created!'

        logger = generate_logger(logname='odf_'+self.odfid, log_dir=self.data_dir)
        logger.log('info', f'Data directory = {self.data_dir}')
        if len(logcomment) > 1:
            logger.log('warning', logcomment)

        # Deal with the rest of the inputs.
        self.level = level
        self.repo = repo

        # Checking LHEASOFT, SAS_DIR and SAS_CCFPATH

        lheasoft = os.environ.get('LHEASOFT')
        if not lheasoft:
            logger.log('error', 'LHEASOFT is not set. Please initialise HEASOFT')
            raise Exception('LHEASOFT is not set. Please initialise HEASOFT')
        else:
            logger.log('info', f'LHEASOFT = {lheasoft}')

        sasdir = os.environ.get('SAS_DIR')
        if not sasdir:
            logger.log('error', 'SAS_DIR is not defined. Please initialise SAS.')
            raise Exception('SAS_DIR is not defined. Please initialise SAS.')
        else:
            logger.log('info', f'SAS_DIR = {sasdir}') 

        sas_ccfpath = os.environ.get('SAS_CCFPATH')
        if not sas_ccfpath:
            logger.log('error', 'SAS_CCFPATH not set. Please define it.')
            raise Exception('SAS_CCFPATH not set. Please define it.')
        else:
            logger.log('info',f'SAS_CCFPATH = {sas_ccfpath}')
        
        os.chdir(self.data_dir)
        logger.log('info', f'Changed directory to {self.data_dir}')

        print(f'''

        Starting SAS session

        Data directory = {self.data_dir}

        ''')


        self.download_data(repo             = self.repo,
                           level            = self.level,
                           data_dir         = self.data_dir,
                           overwrite        = overwrite,
                           logger           = logger,
                           proprietary      = kwargs.get('proprietary', False),
                           credentials_file = kwargs.get('credentials_file', None),
                           encryption_key   = kwargs.get('encryption_key', None),
                           PPS_subset       = kwargs.get('PPS_subset', False),
                           instname         = kwargs.get('instname', None),
                           expflag          = kwargs.get('expflag', None),
                           expno            = kwargs.get('expno', None),
                           product_type     = kwargs.get('product_type', None),
                           datasubsetno     = kwargs.get('datasubsetno', None),
                           sourceno         = kwargs.get('sourceno', None),
                           extension        = kwargs.get('extension', None),
                           filename         = kwargs.get('filename', None))
        
        if hasattr(self, 'odf_dir'):
            if os.path.exists(self.odf_dir):
                self.files['ODF'] = self.__get_list_of_ODF_files()
        if hasattr(self, 'pps_dir'):
            if os.path.exists(self.pps_dir):
                self.files['PPS'] = self.__get_list_of_PPS_files()

        self.work_dir = os.path.join(self.obs_dir,'work')
        if not os.path.exists(self.work_dir): os.mkdir(self.work_dir)
        
        # If only PPS files requested, no processing needed.
        if self.level == 'PPS': return

        self.calibrate_odf(obs_dir        = self.data_dir,
                           sas_ccf        = kwargs.get('sas_ccf', None),
                           sas_odf        = kwargs.get('sas_odf', None),
                           cifbuild_opts  = kwargs.get('cifbuild_opts', []),
                           odfingest_opts = kwargs.get('odfingest_opts', []),
                           recalibrate    = recalibrate)

        if run_epproc and not run_epchain:
            self.__run_analysis('epproc',
                                kwargs.get('epproc_args', []),
                                rerun   = rerun,
                                logFile = 'epproc.log')
        
        if run_epchain:
            self.__run_analysis('epchain',
                                kwargs.get('epchain_args', []),
                                rerun   = rerun,
                                logFile = 'epchain.log')

        if run_emproc and not run_emchain:
            self.__run_analysis('emproc',
                                kwargs.get('emproc_args', []),
                                rerun   = rerun,
                                logFile = 'emproc.log')
            
        if run_emchain:
            self.__run_analysis('emchain',
                                kwargs.get('emchain_args', []),
                                rerun   = rerun,
                                logFile = 'emchain.log')
            
        if run_rgsproc:
            self.__run_analysis('rgsproc',
                                kwargs.get('rgsproc_args', []),
                                rerun   = rerun,
                                logFile = 'rgsproc.log')
        
        #if run_omichain:
        #    self.__run_analysis('omichain',
        #                        kwargs.get('omichain_args', []),
        #                        rerun   = rerun,
        #                        logFile = 'omichain.log')
        
        return
    
    def download_data(self,repo='esa',level='ODF',
                      data_dir=None,overwrite=False,logger=None,
                      proprietary=False,credentials_file=None,
                      encryption_key=None,PPS_subset=False,
                      instname=None,expflag=None,expno=None,
                      product_type=None,datasubsetno=None,
                      sourceno=None,extension=None,
                      filename=None,
                      **kwargs):
        """
        This handles preliminary setup for downloading data files, then 
        calls download_data (as "dl_data") from sasutils.

        Inputs:
            --REQUIRED--

                NONE

            --OPTIONAL--

            --repo:           (string): Which repository to use to download data. 
                                        Default: 'esa'
                                        Can be either
                                        'esa' (data from Europe/ESA) or 
                                        'heasarc' (data from North America/NASA) or
                                        'sciserver' (if user is on sciserver)

            --level:          (string): Level of data products to download.
                                        Default: 'ODF'
                                        Can be 'ODF, 'PPS' or 'ALL'.

            --data_dir:  (string/path): Path to directory where the data will be 
                                        downloaded. Automatically creates directory
                                        data_dir/odfid.
                                        Default: Default from sas_config file, or
                                        current working directory.

            --overwrite:     (boolean): If True will force overwrite of data if odfid 
                                        data already exists in data_dir/odfid.

            --logger:     (TaskLogger): Only used if called from inside 'basic_setup'.

            --proprietary    (boolean): Flag for downloading proprietary data from
                                        the XSA at ESA.

            --credentials_file (filename): Path and filename of file containing XSA
                                        username and password. For proprietary data
                                        only. (Optinal, astroquery will ask user 
                                        for username and password if filename
                                        not given.)

            --encryption_key: (string): Encryption key for proprietary data, a string 32 
                                        characters long. -OR- path to file containing 
                                        ONLY the encryption key.
                                        Note: ONLY used for data from the HEASARC.

            --PPS_subset:    (boolean): Set PPS_subset=True if downloading a subset of PPS
                                        files form the XMM-Newton archive.

            --filename:       (string): If the exact PPS file name is known, then this can
                                        be used to download a single PPS file.

            
            The remaining inputs are used for downloading groups of PPS files using a 
            particular file pattern. Using these requires an understanding of PPS 
            filenames.
            
                instname: instrument name
                expflag: Exposure flag
                expno: Exposure number
                product_type: Product type
                datasubsetno: data subset number/character
                sourceno: Source number or slew step number
                extension: File format
        """
        
        # Where are we?
        startdir = os.getcwd()

        call_download_data = True

        # Brief check to see if data_dir was 
        # given on odfobject creation.
        if self.data_dir != None:
            data_dir = self.data_dir

        # Start checking data_dir
        if data_dir == None:
            if sas_cfg.config.has_option('sas','data_dir'):
                data_dir = sas_cfg.get_setting('data_dir')
            else:
                data_dir = '/does/not/exist'
            if os.path.exists(data_dir):
                self.data_dir = data_dir
            else:
                self.data_dir = startdir
        else:
            self.data_dir = data_dir
        
        if logger == None:
            logger = generate_logger(logname=f'download_{self.odfid}', log_dir=self.data_dir)

        # Identify the download level.
        levelopts = ['ALL','ODF','PPS','4XMM','om_mosaic']
        if level not in levelopts:
            logger.log('error', 'Obs ID request level is undefined!')
            print(f'Options for level are {levelopts[0]}, {levelopts[1]}, or {levelopts[2]}')
            raise Exception('ODF request level is undefined!')
        
        # Set the obs_dir.
        self.obs_dir = os.path.join(self.data_dir,self.odfid)

        # Set odf_dir and pps_dir
        if level == 'ODF' or level == 'ALL':
            self.odf_dir = os.path.join(self.obs_dir,'ODF')
        if level == 'PPS' or level == 'ALL':
            self.pps_dir = os.path.join(self.obs_dir,'PPS')
        self.level = level

        # Checks if obs_dir exists. 
        # Removes it if overwrite = True. Default overwrite = False.
        if os.path.exists(self.obs_dir):
            logger.log('info', f'Existing directory for {self.odfid} found ...')
            if overwrite:
                # If obs_dir exists and overwrite = True then remove obs_dir.
                logger.log('info', f'Removing existing directory {self.obs_dir} ...')
                print(f'\n\nRemoving existing directory {self.obs_dir} ...')
                shutil.rmtree(self.obs_dir)
            else:
                # Check for files
                what_exists = self.__parse_obs_dir()
                if what_exists['odf_dir'] and what_exists['manifest'] and level == 'ODF':
                    logger.log('info', f'Existing ODF directory {self.odf_dir} found ...')
                    call_download_data = False
                if what_exists['pps_dir'] and what_exists['PPS_files'] and level == 'PPS':
                    logger.log('info', f'Existing PPS directory {self.pps_dir} found ...')
                    call_download_data = False
                if (what_exists['odf_dir'] and what_exists['manifest'] and 
                    what_exists['pps_dir'] and what_exists['PPS_files'] and level == 'ALL'):
                    logger.log('info', f'Existing data directories {self.odf_dir} and {self.pps_dir} found ...')
                    call_download_data = False
                if (filename or PPS_subset) and level == 'PPS':
                    logger.log('info', f'Downloading subset of PPS data. Will silently overwrite any pre-existing files.')
                    call_download_data = True
                if not call_download_data:
                    logger.log('info', f'Data found in {self.obs_dir} not downloading again.')
                    print(f'Data found in {self.obs_dir} not downloading again.')

        # Set work directory.
        self.work_dir = os.path.join(self.obs_dir,'work')

        if call_download_data:
            logger.log('info', f'Will download Obs ID data with level {level}')

            # Check chosen repository.
            repo_opts = ['esa','heasarc','sciserver']
            if repo not in repo_opts:
                logger.log('error', 'Download repository not found!')
                print(f'Options for repo are {repo_opts[0]}, {repo_opts[1]}, or {repo_opts[2]}')
                raise Exception('Download repository not found!')
            else:
                logger.log('info', f'Will download data from {repo}.')

            self.repo = repo

            # Function for downloading a single odfid set.
            dl_data(self.odfid,
                    self.data_dir,
                    level=self.level,
                    encryption_key=encryption_key,
                    repo=self.repo,
                    logger=logger,
                    proprietary=proprietary,
                    credentials_file=credentials_file,
                    overwrite=overwrite,
                    PPS_subset=PPS_subset,
                    instname=instname,
                    expflag=expflag,
                    expno=expno,
                    product_type=product_type,
                    datasubsetno=datasubsetno,
                    sourceno=sourceno,
                    extension=extension,
                    filename=filename,
                    **kwargs)
            
        logger.log('info', f'Data directory: {self.data_dir}')
        logger.log('info', f'ObsID directory: {self.obs_dir}')
        print(f'Data directory: {self.data_dir}')
        
        if hasattr(self, 'odf_dir'):
            if os.path.exists(self.odf_dir):
                self.files['ODF'] = self.__get_list_of_ODF_files()
        if hasattr(self, 'pps_dir'):
            if os.path.exists(self.pps_dir):
                self.files['PPS'] = self.__get_list_of_PPS_files()

        return
    
    def calibrate_odf(self,obs_dir=None,
                      sas_ccf=None,sas_odf=None,
                      cifbuild_opts=[],odfingest_opts=[],
                      recalibrate=False,logger=None):
        """
        Before running this function an ODFobject must be created first. e.g.

            odf = pysas.odfcontrol.ODFobject(obsid)

        *Then* the data must be downloaded using:

            odf.download_data()

        This function can then be used as, 
        
            odf.calibrate_odf()

        The calibrate_odf function will automatically look in data_dir for the subdirectory 
        data_dir/odfid. If it does not exist then it will download the data.
        
        If it exists it will search data_dir/odfid and any subdirectories for the ccf.cif
        and *SUM.SAS files. But if overwrite=True then it will remove data_dir/odfid and 
        download the data.

        Optionally the paths to the ccf.cif and *SUM.SAS files can be given through 
        sas_ccf and sas_odf respectively.

        Inputs:
            --REQUIRED--

                NONE

            --OPTIONAL--

            --obs_dir:  (string/path): Path to the obs directory. If no path 
                                       given, then will look in 
                                       data_dir/odfid/. If directory exists then 
                                       will look for ccf.cif and *SUM.SAS files. 
                                       Default: None, looks in data_dir/odfid/.

            --sas_ccf:   (string/path): Path to ccf.cif file for odfid.

            --sas_odf:   (string/path): Path to *SUM.SAS file for odfid.

            --cifbuild_opts:    (list): Options for cifbuild.

            --odfingest_opts:   (list): Options for odfingest.

            --recalibrate:   (boolean): If True will rerun odfingest and cifbuild.

            --logger:     (TaskLogger): Only used if called from inside 'basic_setup'.

        """

        # Where are we?
        startdir = os.getcwd()

        # Brief check to see if data_dir is already defined.
        if self.data_dir == None:
            if sas_cfg.config.has_option('sas','data_dir'):
                data_dir = sas_cfg.get_setting('data_dir')
            else:
                data_dir = '/does/not/exist'
            if os.path.exists(data_dir):
                self.data_dir = data_dir
            else:
                self.data_dir = startdir

        if obs_dir == None:
            self.obs_dir = os.path.join(self.data_dir,self.odfid)

        if logger == None:
            logger = generate_logger(logname='odf_'+self.odfid, log_dir=self.data_dir)

        # Check if obs_dir exists. If not then raise an Exception.
        if not os.path.isdir(self.obs_dir):
            logger.log('error', 'Observation directory: {self.obs_dir} does not exist!')
            print(f'Error! Observation directory: {self.obs_dir} does not exist!')
            print(f'Please provide the path to the observation directory \n \
                    using the input obs_dir=path/to/obs/dir/.')
            raise Exception(f'Error! Observation directory: {self.obs_dir} does not exist!')

        logger.log('info', f'Observation directory = {self.obs_dir}')

        # Deal with the rest of the inputs.
        self.files['sas_ccf'] = sas_ccf
        self.files['sas_odf'] = sas_odf
        if cifbuild_opts == None:
            cifbuild_opts = []
        self.cifbuild_opts = cifbuild_opts
        if odfingest_opts == None:
            odfingest_opts = []
        self.odfingest_opts = odfingest_opts
        
        os.chdir(self.obs_dir)
        logger.log('info', f'Changed directory to {self.obs_dir}')

        print(f'''

        Starting SAS session

        Data directory = {self.obs_dir}

        ''')

        # Set directories for the observation, odf, pps, and work.
        self.odf_dir  = os.path.join(self.obs_dir,'ODF')
        self.work_dir = os.path.join(self.obs_dir,'work')

        # Check what exists in the obs_dir.
        what_exists = self.__parse_obs_dir()

        # Runs calibration if recalibrate = True. Default recalibrate = False
        # Else, looks for ccf.cif and *SUM.SAS files.
        # If ccf.cif and *SUM.SAS files are not found then will run calibration.
        
        if recalibrate:
            if what_exists['odf_dir'] and what_exists['ODF_files']:
                self.__run_calibration(cifbuild_opts,odfingest_opts,logger)
            else:
                logger.log('error', 'ODF directory and files not found!')
                print('ODF directory and files not found! Try downloading data again.')
                raise Exception('ODF directory and files not found!')
            return
        else:
            logger.log('info', f'Searching {self.obs_dir} for ccf.cif and *SUM.SAS files ...')
            # Looking for ccf.cif file.
            if self.files['sas_ccf'] == None:
                ccf_exists = self.get_ccf_cif(logger)
            else:
                # Check if ccf.cif file path given by user exists.
                try:
                    os.path.exists(self.files['sas_ccf'])
                    logger.log('info', '{0} is present'.format(self.files['sas_ccf']))
                    ccf_exists = True
                except FileExistsError:
                    # The only way to get this error is if the user provided a bad filename or path.
                    logger.log('error', 'File {0} not present! Please check if path is correct!'.format(self.files['sas_ccf']))
                    print('File {0} not present! Please check if path is correct!'.format(self.files['sas_ccf']))
                    sys.exit(1)
            
            # Looking for *SUM.SAS file.
            if self.files['sas_odf'] == None:
                SUM_exists = self.get_SUM_SAS(logger)                    
            else:
                # Check if *SUM.SAS file path given by user exists.
                try:
                    SUM_exists = self.get_SUM_SAS(logger,user_defined_file=self.files['sas_odf'])
                    if SUM_exists:
                        logger.log('info', '{0} is present'.format(self.files['sas_odf']))
                except FileExistsError:
                    # The only way to get this error is if the user provided a bad filename or path.
                    logger.log('error', 'File {0} not present! Please check if path is correct!'.format(self.files['sas_odf']))
                    print('File {0} not present! Please check if path is correct!'.format(self.files['sas_odf']))
                    sys.exit(1)

            if ccf_exists and SUM_exists:
                # Set 'SAS_CCF' enviroment variable.
                os.environ['SAS_CCF'] = self.files['sas_ccf']
                logger.log('info', 'SAS_CCF = {0}'.format(self.files['sas_ccf']))
                print('SAS_CCF = {}'.format(self.files['sas_ccf']))

                # Set 'SAS_ODF' enviroment variable.
                os.environ['SAS_ODF'] = self.files['sas_odf']
                logger.log('info', 'SAS_ODF = {0}'.format(self.files['sas_odf']))
                print('SAS_ODF = {0}'.format(self.files['sas_odf']))
            else:
                # If the ccf.cif or *SUM.SAS files are not present, then run calibration.
                self.__run_calibration(cifbuild_opts,odfingest_opts,logger) 
            
            # Set 'SAS_ODF' enviroment variable.
            os.environ['SAS_ODF'] = self.files['sas_odf']
            logger.log('info', 'SAS_ODF = {0}'.format(self.files['sas_odf']))
            print('SAS_ODF = {0}'.format(self.files['sas_odf']))

            self.get_active_instruments()

            if not os.path.exists(self.work_dir): os.mkdir(self.work_dir)
            # Exit the calibrate_odf function. Everything is set.
        
        if os.path.exists(self.odf_dir):
            self.files['ODF'] = self.__get_list_of_ODF_files()

        return
    
    def get_active_instruments(self):
        """
        Checks odf summary file for which instruments were active for that odf.

        Assumes that 'sas_odf' already exists and contains the correct path.

        Also assumes file name and path are stored in self.files['sas_odf'].
        """

        # Get information about the instruments.
        self.active_instruments = {}
        with open(self.files['sas_odf']) as inf:
            lines = inf.readlines()
            for i,line in enumerate(lines):
                if '// Instrument Record' in line:
                    active = lines[i+4][0]
                    if active == 'N': active = False
                    if active == 'Y': active = True
                    self.active_instruments[lines[i+3][0:2]] = active

        # Basic sanity checks
        bad_sum_file = False
        inst_list = list(self.active_instruments.keys())
        true_list = ['M1', 'M2', 'R1', 'R2', 'PN', 'OM']
        diff = set(inst_list) ^ set(true_list)
        if len(diff) > 0: bad_sum_file = True

        if bad_sum_file:
            print('Something is wrong with the odf summary file: {0}'.format(self.files['sas_odf']))

        return
    
    def inisas(self,sas_dir,sas_ccfpath,verbosity=4,suppress_warning=1):
        """
        --Not intended to be used by the end user. Internal use only.--

        Simple wrapper for 'initializesas' defined in init_sas.py.

        SAS initialization should happen automatically.
        """
        self.sas_dir = sas_dir
        self.sas_ccfpath = sas_ccfpath
        self.verbosity = verbosity
        self.suppress_warning = suppress_warning

        return_info = initializesas(self.sas_dir, self.sas_ccfpath, 
                                    verbosity = self.verbosity, 
                                    suppress_warning = self.suppress_warning)
        print(return_info)

    def sas_talk(self,verbosity=4,suppress_warning=1):
        """
        Simple function to set general SAS veriables 'verbosity' 
        and 'suppress_warning'.
        """

        self.verbosity = verbosity
        self.suppress_warning = suppress_warning

        os.environ['SAS_VERBOSITY'] = '{}'.format(self.verbosity)
        os.environ['SAS_SUPPRESS_WARNING'] = '{}'.format(self.suppress_warning)
    
    def find_event_list_files(self,print_output=True):
        """
        Checks the observation directory (obs_dir) for basic unfiltered 
        event list files created by 'epproc', 'emproc', and 'rgsproc'. 
        Stores paths and file names in self.files.

        'self.files' is a dictionary with the following keys:

            'PNevt_list'
            'M1evt_list'
            'M2evt_list'
            'R1evt_list'
            'R2evt_list'
        """

        self.get_active_instruments()

        # Check if events lists have already been made from the odf files.
        # Have dummy value for the OM because it should be handled 
        # differently. But if not included here then too many checks 
        # will be needed later on.
        inst_list = list(self.active_instruments.keys())
        evt_list_list = {'PN': 'PNevt_list',
                         'M1': 'M1evt_list',
                         'M2': 'M2evt_list',
                         'R1': 'R1evt_list',
                         'R2': 'R2evt_list',
                         'OM': 'XXXXXXXXXX'}
        find_list =     {'PN': 'EPN',
                         'M1': 'EMOS1',
                         'M2': 'EMOS2',
                         'R1': 'R1',
                         'R2': 'R2',
                         'OM': 'OM'}
        inst_name =     {'PN': 'EPIC-pn',
                         'M1': 'EPIC-MOS1',
                         'M2': 'EPIC-MOS2',
                         'R1': 'RGS1',
                         'R2': 'RGS2',
                         'OM': 'OM'}
        
        for item in inst_list: self.files[evt_list_list[item]] = []

        for inst in inst_list:
            exists = False
            # Checking for EPIC event lists.
            files = glob.glob(self.obs_dir+'/**/*.ds', recursive=True)
            for filename in files:
                if (filename.find(find_list[inst]) != -1) and filename.endswith('Evts.ds'):
                    self.files[evt_list_list[inst]].append(os.path.abspath(filename))
                    exists = True
            # Checking for RGS event lists.
            files = glob.glob(self.obs_dir+'/**/*EVENLI*FIT', recursive=True)
            for filename in files:
                if (filename.find(find_list[inst]) != -1):
                    self.files[evt_list_list[inst]].append(os.path.abspath(filename))
                    exists = True
            # # Need to do something different for optical monitor images.
            # # Checking for OM sky aligned image.
            # files = glob.glob(self.obs_dir+'/**/*SIMAGE*FIT', recursive=True)
            # for filename in files:
            #     if (filename.find(find_list[inst]) != -1):
            #         self.files[evt_list_list[inst]].append(os.path.abspath(filename))
            #         exists = True
            # # Checking for OM full-frame sky image.
            # files = glob.glob(self.obs_dir+'/**/*FSIMAG*FIT', recursive=True)
            # for filename in files:
            #     if (filename.find(find_list[inst]) != -1):
            #         self.files[evt_list_list[inst]].append(os.path.abspath(filename))
            #         exists = True
            # # Checking for OM sky image mosaic.
            # files = glob.glob(self.obs_dir+'/**/*RSIMAG*FIT', recursive=True)
            # for filename in files:
            #     if (filename.find(find_list[inst]) != -1):
            #         self.files[evt_list_list[inst]].append(os.path.abspath(filename))
            #         exists = True
            if exists:
                self.files[evt_list_list[inst]].sort()
                if print_output:
                    print(" > {0} {1} event list(s) found.\n".format(len(self.files[evt_list_list[inst]]),inst_name[inst]))
                    for x in self.files[evt_list_list[inst]]:
                        print("    " + x + "\n")
            
        return
    
    def find_rgs_spectra_files(self,print_output=True):
        """
        --Not intended to be used by the end user. Internal use only.--

        Check for RGS spectra files created by rgsproc. Adds them to 
        'files' dictrionary with the keys:

            'R1SPEC'
            'R2SPEC'
        """

        self.get_active_instruments()

        # Check if events lists have already been made from the odf files.

        inst_list = list(self.active_instruments.keys())
        rgs_list = []
        if 'R1' in inst_list: rgs_list.append('R1')
        if 'R2' in inst_list: rgs_list.append('R2')

        dict_key  = {'R1': 'R1spectra',
                     'R2': 'R2spectra'}
        file_key  = {'R1': 'R1',
                     'R2': 'R2'}
        inst_name = {'R1': 'RGS1',
                     'R2': 'RGS2'}
        
        for item in rgs_list: self.files[dict_key[item]] = []

        for inst in rgs_list:
            exists = False
            # Checking for RGS spectra.
            files = glob.glob(self.obs_dir+'/**/*RSPEC*FIT', recursive=True)
            for filename in files:
                if (filename.find(file_key[inst]) != -1):
                    self.files[dict_key[inst]].append(os.path.abspath(filename))
                    exists = True
            if exists:
                self.files[dict_key[inst]].sort()
                if print_output:
                    print(" > {0} {1} spectra found.\n".format(len(self.files[dict_key[inst]]),inst_name[inst]))
                    for x in self.files[dict_key[inst]]:
                        print("    " + x + "\n")

        return
    
    def get_ccf_cif(self,logger):
        """
        --Not intended to be used by the end user. Internal use only.--

        Checks for the ccf.cif file. If it exists, inserts file name in 
        'files' dict.
        """

        exists = False

        # Looking for ccf.cif file.
        self.files['sas_ccf'] = os.path.join('does','not','exist')
        logger.log('info', f'Searching for ccf.cif.')
        for path, directories, files in os.walk(self.obs_dir):
            for file in files:
                if 'ccf.cif' in file:
                    logger.log('info', f'Found ccf.cif file in {path}.')
                    self.files['sas_ccf'] = os.path.join(path,file)
        
        # Check if ccf.cif file exists.
        if os.path.exists(self.files['sas_ccf']):
            logger.log('info', '{0} is present'.format(self.files['sas_ccf']))
            exists = True
        else:
            logger.log('warning', 'ccf.cif file not present! User must run calibrate_odf!')

        return exists
    
    def get_SUM_SAS(self,logger,user_defined_file=None):
        """
        --Not intended to be used by the end user. Internal use only.--

        Checks for the *SUM.SAS file. Making this a function since it is 
        used in several places.
        """

        exists = False

        if user_defined_file == None:
            # Looking for *SUM.SAS file.
            self.files['sas_odf'] = os.path.join('does','not','exist')
            logger.log('info', f'Path to *SUM.SAS file not given. Will search for it.')
            for path, directories, files in os.walk(self.obs_dir):
                for file in files:
                    if 'SUM.SAS' in file:
                        logger.log('info', f'Found *SUM.SAS file in {path}.')
                        self.files['sas_odf'] = os.path.join(path,file)
        # Check if *SUM.SAS file exists.
        if os.path.exists(self.files['sas_odf']):
            logger.log('info', '{0} is present'.format(self.files['sas_odf']))
            exists = True
        else:
            logger.log('warning', 'sas_odf file not present! User must run calibrate_odf!')
            return exists
        
        # Check that the SUM.SAS file PATH keyword points to a real ODF directory
        with open(self.files['sas_odf']) as inf:
            lines = inf.readlines()
            for line in lines:
                if 'PATH' in line:
                    key, path = line.split()
                    if not os.path.exists(path):
                        logger.log('error', f'Summary file PATH {path} does not exist. Rerun calibrate_odf with overwrite=True.')
                        print(f'\nSummary file PATH {path} does not exist. \n\n>>>>Rerun calibrate_odf with overwrite=True.')
                        exists = False
                    MANIFEST = glob.glob(os.path.join(path, 'MANIFEST*'))
                    if not os.path.exists(MANIFEST[0]):
                        logger.log('error', f'Missing {MANIFEST[0]} file in {path}. Missing ODF components? Rerun calibrate_odf with overwrite=True.')
                        print(f'\nMissing {MANIFEST[0]} file in {path}. Missing ODF components? \n\n>>>>Rerun calibrate_odf with overwrite=True.')
                        exists = False

        return exists
    
    def clear_obs_dir(self):
        """
        Function to remove all files and subdirectories from the obs_dir.
        """
        if os.path.exists(self.obs_dir):
            print(f'\n\nRemoving existing directory {self.obs_dir} ...')
            shutil.rmtree(self.obs_dir)

    def clear_work_dir(self):
        """
        Function to remove all files and subdirectories from the work_dir.
        """
        if os.path.exists(self.work_dir):
            print(f'\n\nRemoving existing directory {self.work_dir} ...')
            shutil.rmtree(self.work_dir)
    
    def __set_obsid(self):
        """
        --Not intended to be used by the end user. Internal use only.--

        Basic method for setting the environment variables for a single 
        'ObsID'.

        Checks for the existence of various directories. If a directory 
        is not found then __set_obsid will stop and use 'return' command.
        Directories that will be checked for (in this order):
            
            data_dir
            obs_dir
            odf_dir -or- pps_dir
            work_dir

        Then checks for the ccf.cif, *SUM.SAS files and event lists.

        Similar to download_data, but will not download any data, 
        or do anything other than link to files and directories. 
        """

        # Where are we?
        startdir = os.getcwd()

        # Brief check to see if data_dir was 
        # given on odfobject creation.
        if self.data_dir != None:
            data_dir = self.data_dir
        else:
            if sas_cfg.config.has_option('sas','data_dir'):
                data_dir = sas_cfg.get_setting('data_dir')
            else:
                data_dir = '/does/not/exist'

        # Start checking data_dir
        if os.path.exists(data_dir):
            self.data_dir = data_dir
        else:
            # Nothing can be done! User must run download_data!
            return

        # Set directories for the observation, odf, pps, and work.
        self.obs_dir  = os.path.join(self.data_dir,self.odfid)
        self.odf_dir  = os.path.join(self.obs_dir,'ODF')
        self.pps_dir  = os.path.join(self.obs_dir,'PPS')
        self.work_dir = os.path.join(self.obs_dir,'work')

        if os.path.exists(self.obs_dir):
            print(f'obs_dir found at {self.obs_dir}.')
        else:
            # obs_dir not found! Nothing can be done!
            # User must run download_data!
            return
        
        if os.path.exists(self.odf_dir):
            print(f'odf_dir found at {self.odf_dir}.')
            os.environ['SAS_ODF'] = self.odf_dir
            if os.path.exists(self.pps_dir):
                print(f'pps_dir found at {self.pps_dir}.')
        else:
            if os.path.exists(self.pps_dir):
                print(f'pps_dir found at {self.pps_dir}.')
            else:
                print(f'ODF and PPS directories not found! User must download data!')
                return
            
        # Get lists of ODF and PPS files.
        if os.path.exists(self.odf_dir):
            self.files['ODF'] = self.__get_list_of_ODF_files()
        if os.path.exists(self.pps_dir):
            self.files['PPS'] = self.__get_list_of_PPS_files()
        
        if os.path.exists(self.work_dir):
            print(f'work_dir found at {self.work_dir}.')
        else:
            print(f'Default work_dir not found! User must create it!')
            return

        # Only generate logger if observation directory exists.
        logger = generate_logger(logname='odf_'+self.odfid, log_dir=self.data_dir)
        logger.log('info', f'Data directory = {self.data_dir}')
        logger.log('info', f'Existing directory for {self.odfid} found ...')
        logger.log('info', f'Searching {self.data_dir}/{self.odfid} for ccf.cif and *SUM.SAS files ...')

        # Looking for ccf.cif file.
        exists = self.get_ccf_cif(logger)
        if not exists:
            print('ccf.cif file not present! User must run calibrate_odf!')
            return

        # Set 'SAS_CCF' enviroment variable.
        os.environ['SAS_CCF'] = self.files['sas_ccf']
        logger.log('info', 'SAS_CCF = {0}'.format(self.files['sas_ccf']))
        print('SAS_CCF = {}'.format(self.files['sas_ccf']))

        # Looking for *SUM.SAS file.
        exists = self.get_SUM_SAS(logger)
        if not exists:
            print('*SUM.SAS file not present! User must run calibrate_odf!')
            return
        
        # Set 'SAS_ODF' enviroment variable.
        os.environ['SAS_ODF'] = self.files['sas_odf']
        logger.log('info', 'SAS_ODF = {0}'.format(self.files['sas_odf']))
        print('SAS_ODF = {0}'.format(self.files['sas_odf']))

        # Check for previously generated event lists.
        self.find_event_list_files()
        self.find_rgs_spectra_files()

        # Change to work directory.
        os.chdir(self.work_dir)

        # Exit the __set_obsid function. Everything is set.
        return
    
    def __parse_obs_dir(self):
        """
        --Not intended to be used by the end user. Internal use only.--

        Parses the obs_dir for what is present.

        Returns a dictionary with information on what exists.
        """
        exists = {}
        items = ['obs_dir','odf_dir','pps_dir','work_dir',
                 'ccfcif','SUMSAS','manifest',
                 'ODF_files','PPS_files']
        
        for item in items: exists[item] = False

        # Set directories for the observation, odf, pps, and work.
        obs_dir  = os.path.join(self.data_dir,self.odfid)
        odf_dir  = os.path.join(obs_dir,'ODF')
        pps_dir  = os.path.join(obs_dir,'PPS')
        work_dir = os.path.join(obs_dir,'work')

        if os.path.exists(obs_dir): exists['obs_dir']  = True
        if os.path.exists(odf_dir): 
            exists['odf_dir']  = True
            # Just checks if files are there. Doesn't check if they
            # are real or correct.
            if len(os.listdir(odf_dir)) != 0:
                exists['ODF_files'] = True
        if os.path.exists(pps_dir):
            exists['pps_dir']  = True
            # Just checks if files are there. Doesn't check if they
            # are real or correct.
            if len(os.listdir(pps_dir)) != 0:
                exists['PPS_files'] = True
        if os.path.exists(work_dir): exists['work_dir'] = True
        exists['ccfcif'] = self.__check_for_ccf_cif()
        exists['SUMSAS'] = self.__check_for_SUM_SAS()
        exists['manifest'] = self.__check_for_manifest()

        return exists

    def __run_calibration(self,cifbuild_opts,odfingest_opts,logger):
        """
        --Not intended to be used by the end user. Internal use only.--

        Making this a separate function since it can be called from different 
        inside the function calibrate_odf. Prevents duplication of code.
        """
        # Run cifbuild and odfingest on the new data.
        os.chdir(self.odf_dir)
        logger.log('info', f'Changed directory to {self.odf_dir}')

        # Checks that the MANIFEST file is there
        MANIFEST = glob.glob('MANIFEST*')
        try:
            os.path.exists(MANIFEST[0])
            logger.log('info', f'File {MANIFEST[0]} exists')
        except FileExistsError:
            logger.log('error', f'File {MANIFEST[0]} not present. Please check ODF!')
            print(f'File {MANIFEST[0]} not present. Please check ODF!')
            sys.exit(1)

        # Here the ODF is fully untarred below odfid subdirectory
        # Now we start preparing the SAS_ODF and SAS_CCF
        logger.log('info', f'Setting SAS_ODF = {self.odf_dir}')
        print(f'\nSetting SAS_ODF = {self.odf_dir}')
        os.environ['SAS_ODF'] = self.odf_dir

        # Change to working directory
        if not os.path.exists(self.work_dir): os.mkdir(self.work_dir)
        os.chdir(self.work_dir)

        # Run cifbuild
        logger.log('info', f'Running cifbuild with inputs: {cifbuild_opts} ...')
        print(f'\nRunning cifbuild with inputs: {cifbuild_opts} ...')
        w('cifbuild',cifbuild_opts).run()
        
        # Check whether ccf.cif is produced or not
        ccfcif = glob.glob('ccf.cif')
        try:
            os.path.exists(ccfcif[0])
            logger.log('info', f'CIF file {ccfcif[0]} created')
        except FileExistsError:
            logger.log('error','The ccf.cif was not produced')
            print('ccf.cif file is not produced')
            sys.exit(1)
        
        # Sets SAS_CCF variable
        fullccfcif = os.path.join(self.work_dir, 'ccf.cif')
        logger.log('info', f'Setting SAS_CCF = {fullccfcif}')
        print(f'\nSetting SAS_CCF = {fullccfcif}')
        os.environ['SAS_CCF'] = fullccfcif
        self.files['sas_ccf'] = fullccfcif

        # Now run odfingest
        logger.log('info', f'Running odfingest with inputs: {odfingest_opts} ...')
        print(f'\nRunning odfingest with inputs: {odfingest_opts} ...')
        w('odfingest',odfingest_opts).run()

        # Check whether the SUM.SAS has been produced or not
        sumsas = glob.glob('*SUM.SAS')
        try:
            os.path.exists(sumsas[0])
            logger.log('info', f'SAS summary file {sumsas[0]} created')
        except FileExistsError:
            logger.log('error','SUM.SAS file was not produced') 
            print('SUM.SAS file was not produced')
            sys.exit(1)
        
        # Set the SAS_ODF to the SUM.SAS file
        fullsumsas = os.path.join(self.work_dir, sumsas[0])
        os.environ['SAS_ODF'] = fullsumsas
        logger.log('info', f'Setting SAS_ODF = {fullsumsas}')
        print(f'\nSetting SAS_ODF = {fullsumsas}')
        self.files['sas_odf'] = fullsumsas
        
        # Check that the SUM.SAS file has the right PATH keyword
        with open(self.files['sas_odf']) as inf:
            lines = inf.readlines()
            for line in lines:
                if 'PATH' in line:
                    key, path = line.split()
                    if path != self.odf_dir:
                        logger.log('error', f'SAS summary file PATH {path} mismatchs {self.odf_dir}')
                        raise Exception(f'SAS summary file PATH {path} mismatchs {self.odf_dir}')
                    else:
                        logger.log('info', f'Summary file PATH keyword matches {self.odf_dir}')
                        print(f'\nWarning: Summary file PATH keyword matches {self.odf_dir}')

        self.get_active_instruments()

        print(f'''\n\n
        SAS_CCF = {self.files['sas_ccf']}
        SAS_ODF = {self.files['sas_odf']}
        \n''')

    def __run_analysis(self,task,inargs,rerun=False,logFile='DEFAULT'):
        """
        --Not intended to be used by the end user. Internal use only.--

        A wrapper for the wrapper. Yes. I know.

        This function is not intended to be used by the end user, but is
        only called by 'basic_setup'.

        This will check if output files are present for the selected SAS task.
        If they are, will not rerun that SAS task unless "rerun=True".

        Lists of output files are stored in the dictionary self.files{}.

        SAS Tasks that it currently works for:
            --epproc
            --epchain (Warning current version of epchain fails in SAS v. 21)
            --emproc
            --emchain (Warning current version of emchain fails in SAS v. 21)
            --rgsproc

        More will be added as needed.
        """
        self.task = task
        self.inargs = inargs
        self.rerun = rerun
        self.logFile = logFile

        print(self.inargs)

        # Make sure we are in the right place!
        if os.path.isdir(self.work_dir):
            os.chdir(self.work_dir)
        else:
            print(f'The directory for the observation ID ({self.odfid}) does not seem to exist!\n    {self.obs_dir}')
            print('Has \'calibrate_odf\' been run?')
            raise Exception(f'Problem with the directory for odfID = {self.odfid}!')
        
        self.find_event_list_files(print_output=False)
        self.find_rgs_spectra_files(print_output=False)

        run_ep   = False
        run_em   = False
        run_rgs  = False

        # Check if 'epproc' or 'epchain' has been run.
        if (self.task == 'epproc' or self.task == 'epchain') and self.active_instruments['PN']:
            if len(self.files['PNevt_list']) == 0 or self.rerun:
                run_ep   = True
            else:
                print(" > " + str(len(self.files['PNevt_list'])) + " EPIC-pn event list found. Not running {self.task} again.\n")
                for x in self.files['PNevt_list']:
                    print("    " + x + "\n")
                print("..... OK")

        # Check if 'emproc' has been run.
        elif self.task == 'emproc' and (self.active_instruments['M1'] or self.active_instruments['M2']):
            if (len(self.files['M1evt_list']) == 0 and len(self.files['M2evt_list']) == 0) or self.rerun:
                run_em   = True
            else:
                print(" > " + str(len(self.files['M1evt_list'])) + " EPIC-MOS1 event list found. Not running emproc again.\n")
                for x in self.files['M1evt_list']:
                    print("    " + x + "\n")
                print(" > " + str(len(self.files['M2evt_list'])) + " EPIC-MOS2 event list found. Not running emproc again.\n")
                for x in self.files['M2evt_list']:
                    print("    " + x + "\n")
                print("..... OK")

        # Check if 'rgsproc' has been run.
        elif self.task == 'rgsproc' and (self.active_instruments['R1'] or self.active_instruments['R2']):
            if (len(self.files['R1evt_list']) == 0 and len(self.files['R2evt_list']) == 0) or self.rerun:
                run_rgs  = True
            else:
                print(" > " + str(len(self.files['R1evt_list'])) + " RGS1 event list found. Not running rgsproc again.\n")
                for x in self.files['R1evt_list']:
                    print("    " + x + "\n")
                print(" > " + str(len(self.files['R2evt_list'])) + " RGS2 event list found. Not running rgsproc again.\n")
                for x in self.files['R2evt_list']:
                    print("    " + x + "\n")
                print("..... OK")
        
        if run_ep or run_em or run_rgs:
            print(f"   SAS command to be executed: {self.task}, with arguments; \n{self.inargs}")
            print(f"Running {self.task} ..... \n")
            w(self.task,self.inargs,logFile=self.logFile).run()      # <<<<< Execute SAS task

        # Check if run sucsessfully
        self.find_event_list_files(print_output=False)
        if (len(self.files['PNevt_list']) == 0) and run_ep:
            print("Something has gone wrong. I cant find any event list files after running epproc. \n")
        if (len(self.files['M1evt_list']) == 0 and len(self.files['M2evt_list']) == 0 and run_em):
            print("Something has gone wrong. I cant find any event list files after running emproc. \n")
        if (len(self.files['R1evt_list']) == 0 and len(self.files['R2evt_list']) == 0 and run_rgs):
            print("Something has gone wrong. I cant find any event list files after running rgsproc. \n")
        self.find_rgs_spectra_files(print_output=False)

    def __check_for_ccf_cif(self):
        """
        --Not intended to be used by the end user. Internal use only.--

        Checks if the ccf.cif file exists.
        """
        exists = False

        # Check if ccf.cif file exists.
        for path, directories, files in os.walk(self.obs_dir):
            for file in files:
                if 'ccf.cif' in file:
                    if os.path.exists(os.path.join(path,file)):
                        exists = True
        return exists
    
    def __check_for_SUM_SAS(self):
        """
        --Not intended to be used by the end user. Internal use only.--

        Checks if the the *SUM.SAS file exists.
        """
        exists = False

        # Looking for *SUM.SAS file.
        for path, directories, files in os.walk(self.obs_dir):
            for file in files:
                if 'SUM.SAS' in file:
                    if os.path.exists(os.path.join(path,file)):
                        exists = True
        return exists
    
    def __check_for_manifest(self):
        """
        Checks if manifest file exists.
        """

        exists = False

        MANIFEST = glob.glob(self.obs_dir+'/**/*MANIFEST*', recursive=True)
        if len(MANIFEST) > 0:
            if os.path.exists(MANIFEST[0]): exists = True
        
        return exists
    
    def __get_list_of_ODF_files(self):
        """
        Returns list of all files in the the ODF directory.
        """
        file_list = []
        if os.path.exists(self.odf_dir):
            file_list = glob.glob(self.odf_dir+'/*')

        return file_list
    
    def __get_list_of_PPS_files(self):
        """
        Returns list of all files in the the PPS directory.
        """
        file_list = []
        if os.path.exists(self.pps_dir):
            file_list = glob.glob(self.pps_dir+'/*')

        return file_list
    


# ESA (C) 2000-2021
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#
"""param.py

Defines class paramXmlInfoReader with methods to parse and extract all
the information included in the task parameter file and build several
data structures which allow to handle them when entered either from the
command line or via an argument list.


Methods (Class and instance):

xmlParser:		Parses the parameter file into a Document object.
printHelp:		Prints a table with the parameters read from xmlFile.
defaultValues: 	Returns a dictionary with the default values for each parameter.

Functions:
att(p):                     Fills in the attributes of parameter p
getsub(p):                  Obtains the sub-parameters of parameter p
el2nam(p, pels):            Gets the name off parameter element p
nam2el(pname, rev_pels):    Gets the element of parameter name pname


Class Instantiation

t = paramXmlInfoReader('epproc')
t.xmlParser()

"""

# Standard library imports
from xml.dom import minidom as md
import os
import sys
import glob
from collections import UserDict

# Third party imports
from beautifultable import BeautifulTable

# Local application imports
from pysas.logger import get_logger


class paramXmlInfoReader:
    """paramXmlInfoReader

    Class which provides several methods
    to read and handle all the information contained in the
    the task parameter file. The file is formatted according
    to the DOM standard.
    Once the information is loaded, it is distributed into
    several data structures to allow proper handling
    of any parameters entered from the command line or through
    an argument list.
    """

    def __init__(self, taskname, logger = None):
        """taskname to be handled."""

        self.taskname = taskname
        self.xmlFile = ''
        if logger is None:
            # By default will only output to terminal
            self.logger = get_logger('paramXmlInfoReader')
        else:
            # Use logger that was passed in
            self.logger = logger

        sas_dir = os.environ['SAS_DIR']
        self.logger.debug(f'SAS_DIR: {sas_dir}')
        parfile = self.taskname + '.par'
        self.logger.debug(f'Parfile name: {parfile}')
        files = glob.glob(os.path.join(sas_dir, 'config', parfile))
        self.logger.debug(f'Parfile location: {files}')
        if files:
            self.xmlFile = files[0]
            self.logger.debug(f'xmlFile: {self.xmlFile}')

        if self.xmlFile == '':
            raise Exception(f'Does not exist any file named {parfile}. Wrong syntax?')

    # xmlParser instance method
    def xmlParser(self):
        """
        The first task of xmlParser is to get the whole XML file as an object 
        of type Document (For a list of all objects, look at the xml.dom documentation).
        Then, it produces a list of Element objects in the paramater file which are 
        identified by the tag 'PARAM'. This is done by means of the Element method
        getElementsByTagName.
        
        
        Outputs:
        - params                  : List of all elements with tag PARAM
        - allparams               : Dictionary { parameter: {attributes} }
        - pels                    : Dictionary { element: pname }
        - rev_pels                : Dictionary { pname : element }
        - parmap                  : Dictionary { pname: [ subparameters ...], ...}
        
        It is the map of the parameter tree as a list of dictionaries.
        Each dictionary corresponds to a paramater and all its sub-parameters
        structure as follows:
        keys are the parent parameters and values are lists with all their subparameters
        
        
        Examples:
        
        1- Parent parameter: keepfilteroutput
           Subparameters: withfilteredset, filteredset

           {'keepfilteroutput': ['withfilteredset, 'filteredset'],
           'wilthfilteredset': ['filteredset'],
           'fileteredset': [] }
           
           This dictionary corresponds to a parameter tree as follows:
           
           keepfilterouput
               └── withfilteredset
                       └── filteredset

        2- Parent parameter: filtertype
           Subparameters: dssblock, expression
           
           {'filtertype': ['dssblock', 'expression'],
           'dssblok': [],
           'expression': [] }
           
           filtertype
               ├── dssblock
               └── expression


        3- Parent parameter: table
           Subparameters: None
           
           {'table': [] }
           
           - mandpar                 : List of all parameters/subparameters defined mandatory = 'yes'.
           - mainparams              : List of main paramaters, with or without subparamaters
           - mandpar_dict            : Dictionary, key = mandatory subparamater, value=its parent parameter
           - rev_mandpar_dict        : Dictionary, keys  = parent paramater, 
                        value = list of all mandatory subparameters
           - rev_mandpar_string_dict : Dictionary, keys  = parent parameter type 'string'
                        value = list of alternatives
        """

        try:
            doc = md.parse(self.xmlFile)
        except:
            self.logger.error(f'ERROR opening par file {self.taskname}.par: {self.xmlFile}')
            sys.exit(1)

        # self.params
        self.params = doc.getElementsByTagName('PARAM')

        # Dictionary allparams
        self.allparams = {}
        for p in self.params:
            attrib = self.att(p)
            pname = attrib['id']
            self.allparams[pname] = attrib

        # pels dictionary

        self.pels = {}
        for p in self.params:
            pname = self.att(p)['id']
            self.pels[p] = pname

        # reverse pels dictionary
        self.rev_pels = {v: k for k, v in self.pels.items()}

        # parmap

        self.parmap = []
        pardict = {}
        for p in self.params:
            pname = self.el2nam(p, self.pels)
            if p.parentNode.nodeName == 'CONFIG':
                pardict = {}
            subp = self.getsub(p)
            for k in subp:
                i = subp.index(k)
                subp[i] = self.el2nam(k, self.pels)
            pardict[pname] = subp
            if p.parentNode.nodeName == 'CONFIG':
                self.parmap.append(pardict)

        # mandpar

        self.mandpar = []
        for p in self.params:
            if self.att(p)['mandatory'] == 'yes':
                self.mandpar.append(self.att(p)['id'])

        # mainparams

        self.mainparams = []
        for e in self.parmap:
            n = 0
            for k in e.keys():
                n += 1
                if n == 1:
                    self.mainparams.append(k)
                else:
                    break

        # mandpar_dict
        self.mandpar_dict = {}
        for m in self.mandpar:
            for d in self.parmap:
                for k, v in d.items():
                    if m == k and m in self.mainparams:
                        self.mandpar_dict[m] = m
                        continue
                    elif m in v:
                        self.mandpar_dict[m] = k

        # rev_mandpar_dict
        self.rev_mandpar_dict = {}
        for k, v in self.mandpar_dict.items():
            self.rev_mandpar_dict[v] = self.rev_mandpar_dict.get(v, []) + [k]

        # rev_mandpar_string_dir
        self.rev_mandpar_string_dict = {}
        for pname in self.rev_mandpar_dict.keys():
            if self.allparams[pname]['type'] != 'string':
                continue
            el = self.rev_pels[pname]
            subp = el.getElementsByTagName('CASE')
            if len(subp) >= 1:
                subp = el.getElementsByTagName('CASE')[0]
            else:
                continue
            alternatives = []
            for i in range(subp.childNodes.length):
                if subp.childNodes.item(i).nodeName == 'ITEM':
                    a = subp.childNodes.item(i).attributes['value'].value
                    alternatives.append(a)
            self.rev_mandpar_string_dict[pname] = alternatives


        # Adding the constraints attribute to the parameters attributes in parmap
        for p in self.params:
            pname = self.el2nam(p, self.pels)
            for k in p.childNodes:
                if k.nodeName == 'CONSTRAINTS':
                    constraints = p.getElementsByTagName('CONSTRAINTS')[0].firstChild.data
                    constraints = constraints.strip('\n')
                    constraints = constraints.strip()
                    self.allparams[pname]['constraints'] = constraints

    # Print table of all parameters
    def printHelp(self):
        """
        Prints a table of all parameters with columns
        """
        table = BeautifulTable()
        table.columns.header = ['name', 'mandatory', 'type', 'default', 'description']
        table.columns.width = [50, 15, 10, 30, 50]
        table.maxwidth = 160
        for a in self.allparams.values():
            table.rows.append([a['id'], a['mandatory'], a['type'], a.get('default', None), a['description']])
        print(table)

    # method defaultValues - Returns defaults
    def defaultValues(self):
        """
        Default values might be of different types with multiple values
        For the time being, we create lists for all of then,
        which can be either single or multiple values.
        """

        defaults = {}
        for p, a in self.allparams.items():
            defaults[p] = a.setdefault('default', '')
        return defaults

    # Static methods used in the class

    @staticmethod
    def att(p):
        """
        att(p) obtains the attributes of parameter p
        """
    
        attrib = {}
        for k, v in p.attributes.items():
            attrib[k] = v
        if 'mandatory' not in attrib.keys():
            attrib['mandatory'] = 'no'
        if 'list' not in attrib.keys():
            attrib['list'] = 'no'
        # All parameters have a DESCRIPTION node after them
        if p.getElementsByTagName('DESCRIPTION').length == 0:
            description = ''
        elif p.getElementsByTagName('DESCRIPTION')[0].firstChild == None:
            description = ''
        else:
            description = p.getElementsByTagName('DESCRIPTION')[0].firstChild.data
            description = description.strip('\n')
            description = description.strip()
        attrib['description'] = description
        return attrib

    @staticmethod
    def getsub(p):
        """
        getsub(p) returns the list of all subparamaters
        from a given parameter p downwards the tree structure
        """
        subp = p.getElementsByTagName('PARAM')
        return subp
    
    @staticmethod
    def nam2el(pname, rev_pels):
        """
        nam2el(pname): returns the parameter element p for a parameter name
        """
        if not rev_pels:
            raise Exception('rev_pels is undefined!')
        el = rev_pels[pname]
        return el
    
    @staticmethod
    def el2nam(p, pels):
        """
        el2nam(p): returns the name of a parameter element p
        """
        if not pels:
            raise Exception('pels is undefined!')
        pname = pels[p]
        return pname

class SASParams(UserDict):
    """
    The class 'SASParams' is a child of 'UserDict', which is 
    like a normal Python dict, but allows for special modifications.

    Upon instatiation the object will create a parallel dict which
    keeps track if the values of the main dict have been modified.

    The end user should interact with the SASParams dict like a normal
    Python dictionary.
    
    The special methods are intended for internal pySAS use only!

    How to use:
        from pysas.param import SASParams
        my_task_inputs = SASParams({})
        my_task_inputs.set_inparams_to_defaults(taskname)
    """
    def __init__(self, *args, **kwargs):
        self.inparams = {}
        super().__init__(*args, **kwargs)
        for key in self.data.keys():
            self.inparams[key] = (self.data[key],False)
    def __setitem__(self, key, value):
        self.inparams[key] = (value,True)
        super().__setitem__(key, value)

    def get_task_defaults(self,taskname):
        """
        Function to get the default parameters for a SAS task.
        """
        t = paramXmlInfoReader(taskname)
        t.xmlParser()
        self.defaults = t.defaultValues()
        self.defaults['options'] = ''

    def get_task_params(self,taskname):
        """
        Function to get ALL the parameter information
        for a SAS task.
        """
        t = paramXmlInfoReader(taskname)
        t.xmlParser()
        self.allparams = t.allparams
        self.mandparams = t.mandpar
        self.mainparams = t.mainparams
        self.parmap = t.parmap
        self.mandpar_dict = t.mandpar_dict
        self.rev_mandpar_dict = t.rev_mandpar_dict
        self.rev_mandpar_string_dict = t.rev_mandpar_string_dict

    def set_inparams_to_defaults(self, taskname):
        """
        Sets special dictionary 'inparams' to defaults without
        setting 'modified' to 'True'.

        Silently overwrites previous values!
        """
        self.get_task_defaults(taskname)
        for k, v in self.defaults.items():
            self.inparams[k] = (v,False)
            super().__setitem__(k, v)

def get_input_params(taskname):
    """
    Function to return dictionary of input parameters for a given 
    SAS task.

        Input:
        - taskname (string)

        Output:
        - Special dictionary of input parameters with defaults.
          The dictionary will be of the type 'SASParams', which
          behaves like a normal dictionary, but keeps track of 
          which parameters have been modified.
        
    """
    
    return_dict = SASParams({})
    return_dict.set_inparams_to_defaults(taskname)

    return return_dict
